'----------------------------------------------------------
'	SR2 Chunks Texture Replacement Tool
'	Copyright (c) 2020 Masamichi Kirikae
'----------------------------------------------------------
#include once "SR2ChunksTextureTool.bi"

function AboutDlgProc(byval hWin as HWND,byval uMsg as UINT,byval wParam as WPARAM,byval lParam as LPARAM) as s32
	dim as s32 id,event
	dim lf as LOGFONT

	select case uMsg
		case WM_INITDIALOG
			' Create a back brush
		case WM_COMMAND
			id=loword(wParam)
			event=hiword(wParam)
			if event=BN_CLICKED then
				if id=IDOK then
					SendMessage(hWin,WM_CLOSE,NULL,NULL)
				endif
			endif
			'
		case WM_CLOSE
			EndDialog(hWin,0)
			'
		case else
			return FALSE
			'
	end select
	return TRUE

end function

Function DialogProc(ByVal hWin As HWND, ByVal uMsg As UINT, ByVal wParam As WPARAM, ByVal lParam As LPARAM) As s32
	dim as s32 id, nEvent, flag, ofnFlags, ln, lRes, retval, attrib_mask
	dim as u32 n
	dim as zstring * 260 ztemp, buff
	dim as any ptr thread
'	static as HMENU hMenu
	static as s32 minval
	dim as string path
	dim as string sourcepath, destpath
	static ERROR_STAT as u32

	Select Case uMsg
		Case WM_INITDIALOG
			
			InitCommonControls()
			
			DisableProcessWindowsGhosting()
		
			_hDlg = hWin
			
			'	load Icon
			hIcon = LoadIcon(hInstance, cast(zstring ptr, 500))
			sendmessage(hWin, WM_SETICON, NULL, cast(LPARAM, hIcon))
			
			'	initialize dds instructor
			dds.init_dds()
			'	initialize peg instructor
			peg_pc.init_peg()

			'	read sr2pegtool.ini
			destpath = _INIFILE_NAME
			retval = read_inifile(destpath)
			if (retval = F_ERROR) then
				ERROR_STAT = ERROR_STAT or ERROR_WRONG_INIFILE
			endif
			
			if (sr2path = "") then
				ERROR_STAT = ERROR_STAT or ERROR_NO_SR2FOLDER
			else
				if (dir(sr2path + "\chunks1.vpp_pc", fbArchive) = "") then
					sr2path = ""
					ERROR_STAT = ERROR_STAT or ERROR_NO_SR2FOLDER
				endif				
			endif
		
			'	check temporary folder setting
			retval = init_temporary_folder()
			if (retval = F_ERROR) then
				ERROR_STAT = ERROR_STAT or ERROR_NO_TEMPFOLDER
			endif
			
			'	check original_vpp_pc
			retval = check_vpp_pc_files(_ORIGINAL_VPP_PC)
			if (retval = F_ERROR) then
				ERROR_STAT = ERROR_STAT or ERROR_NO_VPPSOURCE
			endif
			
			'	get chunk number and names
			total_chunks = get_chunk_names()
			if (total_chunks <> _TOTAL_CHUNKS) then
				ERROR_STAT = ERROR_STAT or ERROR_NO_CHUNKSDATA
			endif

			'	init progress bar
			hProg = GetDlgItem(hWin, IDC_PGB1)
			Prog_Val = 0
			SendMessage(hProg, PBM_SETRANGE, 0, MAKELPARAM( 0, 100 ))
			SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
					
			'	get menuhandle
			hMenu = LoadMenu(GetModuleHandle(NULL), MAKEINTRESOURCE(IDR_MENU1))
			retval = SetMenu(hWin, hMenu)
			
			'	List Error Messages
			flag = FALSE
			n = 1
			do while (n < ERROR_MAXVAL)
				retval = ERROR_STAT and n
				if (retval <> 0) then
					flag = TRUE
					select case n
						case ERROR_WRONG_INIFILE
							ztemp = "Wrong ini file description."
							AddMessage(ztemp)
						case ERROR_NO_SR2FOLDER
							ztemp = "SR2 installed folder not set."
							AddMessage(ztemp)							
						case ERROR_NO_TEMPFOLDER
							ztemp = "Temporary folder not set."
							AddMessage(ztemp)
						case ERROR_NO_VPPSOURCE
							ztemp = "No original vpp_pc files."
							AddMessage(ztemp)
						case ERROR_NO_CHUNKSDATA
							ztemp = "No chunks texture data."
							AddMessage(ztemp)
						case else
					end select
				endif
				n*=2
			loop
			
			'	Redraw	Menu	
			if (ERROR_STAT and ERROR_WRONG_INIFILE) then
				MessageBox(hWin, "Correct ini file description.", @"Error", MB_OK or MB_APPLMODAL)
				'DeleteObject(hFont)
				EndDialog(hWin, 0)
			endif
				
			if (ERROR_STAT and ERROR_NO_SR2FOLDER) then
				EnableMenuItem(hMenu, IDM_FILE_COPY, MF_GRAYED)
				DrawMenuBar(hMenu)
			endif
			
			if (ERROR_STAT and ERROR_NO_TEMPFOLDER) then
				EnableMenuItem(hMenu, IDM_MAKE_EXTRACT, MF_GRAYED)
				EnableMenuItem(hMenu, IDM_MAKE_REBUILD, MF_GRAYED)
				DrawMenuBar(hMenu)
			endif

			if (ERROR_STAT and ERROR_NO_VPPSOURCE) then
				EnableMenuItem(hMenu, IDM_MAKE_EXTRACT, MF_GRAYED)
				DrawMenuBar(hMenu)
			endif

			if (ERROR_STAT and ERROR_NO_CHUNKSDATA) then
				EnableMenuItem(hMenu, IDM_MAKE_REBUILD, MF_GRAYED)
				EnableMenuItem(hMenu, IDM_MAKE_STAT, MF_GRAYED)
				EnableMenuItem(hMenu, IDM_FILE_INIT, MF_GRAYED)
				DrawMenuBar(hMenu)				
			endif
			
			if (ERROR_STAT = 0) then
				ztemp = "Ready."	'	no error
				AddMessage(ztemp)
			endif

		case WM_COMMAND

			id = LoWord(wParam)
			nEvent = HiWord(wParam)


			' Menu, toolbar and accelerator events
			select case nEvent
				case BN_CLICKED, 1

					if (THREAD_STATUS <> THREAD_NOT_IN_USE) then
						if (THREAD_STATUS = IDM_MAKE_REBUILD) then
							if (id <> IDM_MAKE_ABORT) then
								exit select
							end if
						endif				
					endif

					select case id
						case IDM_FILE_OPEN	' select SR2 installed folder
								
							if (sr2path = "") then
								if (dir("C:\Program Files (x86)\Steam\steamapps\common\Saints Row 2", fbDirectory) <> "") then
									ztemp = "C:\Program Files (x86)\Steam\steamapps\common\Saints Row 2"
								elseif (dir("D:\SteamLibrary\steamapps\common\Saints Row 2", fbDirectory) <> "") then
									ztemp = "D:\SteamLibrary\steamapps\common\Saints Row 2"
								elseif (dir("E:\SteamLibrary\steamapps\common\Saints Row 2", fbDirectory) <> "") then
									ztemp = "E:\SteamLibrary\steamapps\common\Saints Row 2"
								else
									ztemp = currpath
								endif
							else
								ztemp = sr2path
							endif
							
							ofnFlags = BIF_NEWDIALOGSTYLE or BIF_RETURNONLYFSDIRS
							buff = FileSelectFolder("Select SR2 Installed Folder", 0, ofnFlags, ztemp)
							
							if (buff <> "") then
								sourcepath = buff + "\chunks1.vpp_pc"
								if (dir(sourcepath, fbArchive) = "") then
									MessageBox(hWin, "chunks1.vpp_pc not found. Select correct folder.", @"Error", MB_OK or MB_APPLMODAL)
									SR2path = ""
									exit select
								else
									'	set valid SR2 installed folder
									sr2path = buff
									write_inifile(_INIFILE_NAME)
									ERROR_STAT = ERROR_STAT and (not ERROR_NO_SR2FOLDER)
									ztemp = "SR2 installed folder not set."
									DelMessage(ztemp)
									EnableMenuItem(hMenu, IDM_FILE_COPY, MF_ENABLED)
									DrawMenuBar(hMenu)
									ztemp = "SR2 installed folder is set to" + !"\r\n" + sr2path
									MessageBox(hWin, ztemp, @"Info", MB_OK or MB_APPLMODAL)
								endif
							endif
							
							if ((ERROR_STAT and ERROR_NO_VPPSOURCE) <> 0) then
								MessageBox(hWin, "Copy original chunks1_vpp_pc, chunks2.vpp_pc, chunks3.vpp_pc and pegs.vpp_pc.", @"Info", MB_OK or MB_APPLMODAL)

								if ( check_vpp_pc_files(_ORIGINAL_VPP_PC) = F_ERROR ) then
									ERROR_STAT = ERROR_STAT and (not ERROR_NO_VPPSOURCE)
									dim as any ptr handle
									handle = ThreadCreate(@copy_vpp_pc_sourcefiles)
									ThreadDetach(handle)
								endif
								
								ztemp = "No original vpp_pc files."
								DelMessage(ztemp)
									
							else
								MessageBox(hWin, "The original vpp_pc files have already been copied.", @"Error", MB_OK or MB_APPLMODAL)
							endif

							'	if valid temp folder was set
							if ((ERROR_STAT and ERROR_NO_TEMPFOLDER) = 0) then
								EnableMenuItem(hMenu, IDM_MAKE_EXTRACT, MF_ENABLED)
								if ((ERROR_STAT and ERROR_NO_CHUNKSDATA) = 0) then
									EnableMenuItem(hMenu, IDM_MAKE_REBUILD, MF_ENABLED)
									EnableMenuItem(hMenu, IDM_MAKE_STAT, MF_ENABLED)
									EnableMenuItem(hMenu, IDM_FILE_INIT, MF_ENABLED)
									DrawMenuBar(hMenu)
									if (ERROR_STAT = 0) then
										DelMessage("Ready.")
										AddMessage("Ready.")
									endif
								endif
							endif							
								
						case IDM_FILE_COPY
							
							'	check new_vpp_pc
							retval = check_vpp_pc_files(_REBUILD_VPP_PC)
							if (retval = F_ERROR) then exit select

							ztemp = "Copy new chunks1,2,3 and pegs.vpp_pc files to" + !"\r\n" + sr2path + "?"

							retval = MessageBox(hWin, ztemp, @"Info", MB_YESNO or MB_APPLMODAL)
							if (retval <> IDYES) then exit select

							dim as any ptr handle
							handle = ThreadCreate(@copy_vpp_pc_newfiles)
							ThreadDetach(handle)
							
						
						case IDM_FILE_INIT
							
							ztemp = "Delete all chunks texture data?"

							retval = MessageBox(hWin, ztemp, @"Info", MB_YESNO or MB_APPLMODAL)
							if (retval <> IDYES) then exit select
							
							dim as any ptr handle
							handle = ThreadCreate(@init_chunks_data)
							ThreadDetach(handle)
							
							ERROR_STAT = ERROR_STAT or ERROR_NO_CHUNKSDATA
							EnableMenuItem(hMenu, IDM_MAKE_REBUILD, MF_GRAYED)
							EnableMenuItem(hMenu, IDM_MAKE_STAT, MF_GRAYED)
							EnableMenuItem(hMenu, IDM_FILE_INIT, MF_GRAYED)							
							DrawMenuBar(hMenu)
							DelMessage("Ready.")
							AddMessage("No chunks texture data.")
							
								
						case IDM_FILE_EXIT
							
							delete_temporary_subfolders()
							SendMessage(hWin, WM_CLOSE, 0, 0)
							
						case IDM_EDIT_TEMP
							
							ztemp = currpath
							ofnFlags = BIF_NEWDIALOGSTYLE or BIF_RETURNONLYFSDIRS
							buff = FileSelectFolder("Select Temporary Folder", 0, ofnFlags, ztemp)
							if (buff <> "") then
								'	check temporary folder setting
								temppath = buff
								retval = init_temporary_folder()						
								if (retval = F_ERROR) then
									'	invald temp folder
									ERROR_STAT = ERROR_STAT or ERROR_NO_TEMPFOLDER
									MessageBox(hWin, "Temporary folder " + buff + " unavailavle.", @"Info", MB_OK or MB_APPLMODAL)
								else
									'	valid temp folder
									scope
										dim as s32 lg1, lg2
	
										lg1 = len(temppath)
										lg2 = len(rootpath)
										if (lg1 > lg2) then
											if (left(temppath, lg2) = rootpath) then
												temppath = right(temppath, lg1 - lg2 - 1)
											endif
										endif
									end scope
									
									MessageBox(hWin, "Temporary folder set to " + temppath , @"Info", MB_OK or MB_APPLMODAL)
									write_inifile(_INIFILE_NAME)

									if ((ERROR_STAT and ERROR_NO_TEMPFOLDER) <> 0) then
										DelMessage("Temporary folder not set.")
										ERROR_STAT = ERROR_STAT and (not ERROR_NO_TEMPFOLDER)
										if ((ERROR_STAT and ERROR_NO_VPPSOURCE) = 0) then
											EnableMenuItem(hMenu, IDM_MAKE_EXTRACT, MF_ENABLED)
											DrawMenuBar(hMenu)
										endif
										if ((ERROR_STAT and ERROR_NO_CHUNKSDATA) = 0) then
											EnableMenuItem(hMenu, IDM_MAKE_REBUILD, MF_ENABLED)
											EnableMenuItem(hMenu, IDM_MAKE_STAT, MF_ENABLED)
											DrawMenuBar(hMenu)
										endif
										if (ERROR_STAT = 0) then
											DelMessage("Ready.")
											AddMessage("Ready.")
										endif
									endif
								endif
							else
								'	return path is NULL
								MessageBox(hWin, "Invalid Temporay folder path", @"Info", MB_OK or MB_APPLMODAL)
							endif
	
						case IDM_MAKE_EXTRACT
							
							ztemp = "Start the extract chunks texture data?" + !"\r\n" + "This process takes a long time."

							retval = MessageBox(hWin, ztemp, @"Extract chunks texture data", MB_YESNO or MB_APPLMODAL)
							if (retval <> IDYES) then exit select
							
							dim as any ptr handle
							handle = ThreadCreate(@make_chunks_texture_data)
							ThreadDetach(handle)

							if ((ERROR_STAT and ERROR_NO_CHUNKSDATA) <> 0) then
								ztemp = "No chunks texture data."
								DelMessage(ztemp)
								AddMessage("Ready.")
							endif
							
							ERROR_STAT = ERROR_STAT and (not ERROR_NO_CHUNKSDATA)
							EnableMenuItem(hMenu, IDM_MAKE_REBUILD, MF_ENABLED)
							EnableMenuItem(hMenu, IDM_MAKE_STAT, MF_ENABLED)
							EnableMenuItem(hMenu, IDM_FILE_INIT, MF_ENABLED)
							DrawMenuBar(hMenu)

						case IDM_MAKE_REBUILD
							
							ztemp = "Temporary folder = " + temppath + !"\r\n"
							ztemp += get_dds_settings() + !"\r\n"
							ztemp += "Replace textures and rebuild vpp_pc file?" + !"\r\n" + "This process takes a long time."
							
							retval = MessageBox(hWin, ztemp, @"Relace textures and rebuild", MB_YESNO or MB_APPLMODAL)
							if (retval <> IDYES) then exit select

							EnableMenuItem(hMenu, IDM_MAKE_ABORT, MF_ENABLED)
						
							dim as any ptr handle
							handle = ThreadCreate(@replace_textures_and_rebuild)
							ThreadDetach(handle)
							
							EnableMenuItem(hMenu, IDM_FILE_COPY, MF_ENABLED)
							EnableMenuItem(hMenu, IDM_FILE_INIT, MF_ENABLED)
							DrawMenuBar(hMenu)
							
						case IDM_MAKE_STAT
							
							retval = MessageBox(hWin, "Generate texture_statistics.txt?", @"Info", MB_YESNO or MB_APPLMODAL)
							if (retval<>IDYES) then exit select
							
							dim as any ptr handle
							handle = ThreadCreate(@make_texture_statistics)
							ThreadDetach(handle)
							
						case IDM_MAKE_ABORT

								EnableMenuItem(hMenu, IDM_MAKE_ABORT, MF_GRAYED)
								DrawMenuBar(hMenu)
								THREAD_ABORT = TRUE
	
						case IDM_HELP_ABOUT
							DialogBoxParam(GetModuleHandle(NULL), Cast(zstring ptr,IDD_DLGABOUT), NULL, @AboutDlgProc, NULL)
						case else
					end select
				case else
			end select
			'
		case WM_CLOSE
		
			if (THREAD_STATUS <> THREAD_IN_USE) then
				delete_temporary_subfolders()
				EndDialog(hWin, 0)
			endif
		
		case WM_DESTROY
			
			if (THREAD_STATUS <> THREAD_IN_USE) then
				delete_temporary_subfolders()
				PostQuitMessage(NULL)
			endif
			'
		case WM_SIZE
			'
		case else

			return(FALSE)
			'
	end select

	return(TRUE)

end function

function main() as s32

	currpath = curdir()
	rootpath = exepath
	chdir(rootpath)

	'	Create the Dialog
	hInstance=GetModuleHandle(NULL)
	DialogBoxParam(hInstance,Cast(ZString Ptr,IDD_DLG1),0,@DialogProc,0)

	'	Program has ended
	chdir(currpath)
	ExitProcess(0)
	return(0)
end function

end( main() )

'--------------------------------------------------------------------
'	delete all temporary files
'--------------------------------------------------------------------
sub delete_temporary_files()
	dim as string destpath

	'	cleanup temporary chunks folder
	destpath = temppath + _TEMP_CHUNKS + "\*.*"
	delete_file(destpath, fbArchive)
	'	cleanup temporary dds folder
	destpath = temppath + _TEMP_DDS + "\*.*"
	delete_file(destpath, fbArchive)
	'	cleanup temporary peg folder
	destpath = temppath + _TEMP_PEGS + "\*.*"
	delete_file(destpath, fbArchive)
	
end sub

'----------------------------------------------------------
'	abort thread
'----------------------------------------------------------
sub abort_thread()
	dim as string destpath
	dim as zstring*256 ztemp
	'dim as HMENU hMenu
	'	reset progress bar
	Prog_Val = 0
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
	ztemp = "Aborting Process..."
	AddMessage(ztemp)
	'		cleanup chunks temporary files
	destpath = temppath + _TEMP_CHUNKS + "\*.*"
	delete_file(destpath, FbArchive)
	'		cleanup peg temporary files
	destpath = temppath + _TEMP_PEGS + "\*.*"
	delete_file(destpath, FbArchive)	
	sleep 1000
	'	clear messages
	DelMessage("*")
	'	thread status
	THREAD_STATUS = THREAD_NOT_IN_USE
	THREAD_ABORT = FALSE
	
	MessageBox(_hDlg, "Process aborted.", @"Info", MB_OK or MB_APPLMODAL)

end sub

'--------------------------------------------------------------------
'	delete temporary subfolders
'--------------------------------------------------------------------
sub delete_temporary_subfolders()
	dim as string destpath
	
	'	delete temporary chunks folder
	delete_temporary_files()
	destpath = temppath + _TEMP_CHUNKS
	rmdir(destpath)
	'	delete temporary dds folder
	destpath = temppath + _TEMP_DDS
	rmdir(destpath)
	'	delete temporary peg folder
	destpath = temppath + _TEMP_PEGS
	rmdir(destpath)
end sub

'--------------------------------------------------------------------
'	display message to listbox and stack it
'	when press ESC return TRUE else FALSE
'--------------------------------------------------------------------
sub AddMessage(byref zst as zstring)
	dim as s32 ln, retval
	ln = SendMessage(GetDlgItem(_hDlg, IDC_LST1), LB_ADDSTRING, 0, cast(LPARAM, @zst))
	UpdateWindow(GetDlgItem(_hDlg, IDC_LST1))
	ListStack(StackDepth).n = ln
	ListStack(StackDepth).message = zst
	StackDepth += 1
end sub

'--------------------------------------------------------------------
'	delete stack message from listbox
'--------------------------------------------------------------------
sub DelMessage(byref zst as zstring)
	dim as s32 ln, i
	select case zst
		case ""
			if (StackDepth > 0) then
				StackDepth -= 1
				ln = ListStack(StackDepth).n
				SendMessage(GetDlgItem(_hDlg, IDC_LST1), LB_DELETESTRING, ln, 0)
				UpdateWindow(GetDlgItem(_hDlg, IDC_LST1))
				ListStack(StackDepth).n = 0
				ListStack(StackDepth).message = ""
			endif
		case "*"
			for i = StackDepth - 1 to 1 step -1
				ln = ListStack(i).n
				SendMessage(GetDlgItem(_hDlg, IDC_LST1), LB_DELETESTRING, ln, 0)
				UpdateWindow(GetDlgItem(_hDlg, IDC_LST1))
				ListStack(StackDepth).n = 0
				ListStack(StackDepth).message = ""
			next
		case else
			ln = SendMessage(GetDlgItem(_hDlg, IDC_LST1), LB_FINDSTRING, 0, cast(LPARAM, @zst))
			if (ln <> -1) then
				SendMessage(GetDlgItem(_hDlg, IDC_LST1), LB_DELETESTRING, ln, 0)
				for i = ln to StackDepth-2
					ListStack(i) = listStack(i+1)			
				next
				StackDepth -= 1
				ListStack(StackDepth).n = 0
				ListStack(StackDepth).message = ""			
			endif	
	end select
end sub

'--------------------------------------------------------------------
'	Display Elapsed Time and Remaining Time. Use with Progress Bar
'--------------------------------------------------------------------
sub display_time()
	dim as s32 ln
	static as double start_time
	dim as double percent, current_time, elapsed_time, remain_time
	dim as string elapsed_time_str, remain_time_str
	dim as zstring*260 ztemp

	if (Prog_Val = 0) then
		start_time = timer
		ztemp = "Elapsed Time"
		ln = SendMessage(GetDlgItem(_hDlg, IDC_LST2), LB_GETCOUNT, 0, 0)
		if (ln > 0) then
			SendMessage(GetDlgItem(_hDlg, IDC_LST2), LB_DELETESTRING, 0, 0)
			UpdateWindow(GetDlgItem(_hDlg, IDC_LST2))
		endif
	else
		ln = SendMessage(GetDlgItem(_hDlg, IDC_LST2), LB_GETCOUNT, 0, 0)
		if (ln > 0) then
			SendMessage(GetDlgItem(_hDlg, IDC_LST2), LB_DELETESTRING, 0, 0)
		endif
		percent = Prog_Val/Prog_Max
		current_time = timer
		elapsed_time = (current_time - start_time) * 1.0/percent
		elapsed_time_str = libstr.format_hms(current_time - start_time)
		remain_time = cint(elapsed_time - (current_time - start_time))
		remain_time_str = libstr.format_hms(remain_time)
		ztemp = "Elapsed Time " + elapsed_time_str + "    Remaining Time " + remain_time_str
		SendMessage(GetDlgItem(_hDlg, IDC_LST2), LB_ADDSTRING, 0, cast(LPARAM, @ztemp))
		UpdateWindow(GetDlgItem(_hDlg, IDC_LST2))
	endif
end sub

'--------------------------------------------------------------------
'	when all vpp_pc files exist return F_NOERROR else return F_ERROR
'--------------------------------------------------------------------
function check_vpp_pc_files(path as string) as s32
	dim as string fpath
	
	fpath = path + "\chunks1.vpp_pc"
	if (dir(fpath, fbArchive) = "") then return(F_ERROR)

	fpath = path + "\chunks2.vpp_pc"
	if (dir(fpath, fbArchive) = "") then return(F_ERROR)

	fpath = path + "\chunks3.vpp_pc"
	if (dir(fpath, fbArchive) = "") then return(F_ERROR)

	fpath = path + "\pegs.vpp_pc"
	if (dir(fpath, fbArchive) = "") then return(F_ERROR)

	return(F_NOERROR)	
end function

'--------------------------------------------------------------------
'	copy original *.vpp_pc files to original_vpp_pc folder
'--------------------------------------------------------------------
sub copy_vpp_pc_sourcefiles()
	dim as string sourcepath, destpath
	'dim as zstring*260 ztemp
	dim as s32 retval, ln
	
	THREAD_STATUS = THREAD_IN_USE
	THREAD_ERROR = F_NOERROR

	Prog_Val = 0
	Prog_Max = 3
	SendMessage(hProg, PBM_SETRANGE, 0, MAKELPARAM( 0, Prog_Max ))
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)

	sourcepath = SR2path + "\chunks1.vpp_pc"
	destpath = _ORIGINAL_VPP_PC
	AddMessage("Copying chunks1.vpp_pc...")
	copy_file(sourcepath, destpath, fbArchive)
	DelMessage("")
	
	Prog_Val+=1
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
	
	sourcepath = SR2path + "\chunks2.vpp_pc"
	destpath = _ORIGINAL_VPP_PC
	AddMessage("Copying chunks2.vpp_pc...")
	copy_file(sourcepath, destpath, fbArchive)
	DelMessage("")

	Prog_Val+=1
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)

	sourcepath = SR2path + "\chunks3.vpp_pc"
	destpath = _ORIGINAL_VPP_PC
	AddMessage("Copying chunks3.vpp_pc...")
	copy_file(sourcepath, destpath, fbArchive)
	DelMessage("")
	
	Prog_Val+=1
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
	
	sourcepath = SR2path + "\pegs.vpp_pc"
	destpath = _ORIGINAL_VPP_PC
	AddMessage("Copying pegs.vpp_pc...")
	copy_file(sourcepath, destpath, fbArchive)
	DelMessage("")

	Prog_Val+=1
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
	
	MessageBox(NULL, "Original vpp_pc files was copied to the original_vpp_pc folder.", @"Info", MB_OK)
	
	Prog_Val = 0
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
	
	THREAD_STATUS = THREAD_NOT_IN_USE
	THREAD_RETURN = F_NOERROR
	return
end sub

'--------------------------------------------------------------------
'	copy new *.vpp_pc files to SR2 installed folder
'--------------------------------------------------------------------
sub copy_vpp_pc_newfiles()
	dim as string sourcepath, destpath
	dim as zstring*260 ztemp
	dim as s32 retval, ln
	
	THREAD_STATUS = THREAD_IN_USE
	THREAD_ERROR = F_NOERROR
	
	Prog_Val = 0
	Prog_Max = 3
	SendMessage(hProg, PBM_SETRANGE, 0, MAKELPARAM( 0, Prog_Max ))
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)

	retval = check_vpp_pc_files(_REBUILD_VPP_PC)
	if (retval = TRUE) then
		THREAD_STATUS = THREAD_NOT_IN_USE
		THREAD_RETURN = F_NOERROR
		return
	EndIf
	
	sourcepath = _REBUILD_VPP_PC + "\chunks1.vpp_pc"
	destpath = sr2path
	AddMessage("Copying chunks1.vpp_pc...") 
	copy_file(sourcepath, destpath, fbArchive)
	
	Prog_Val+=1
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
	
	sourcepath = _REBUILD_VPP_PC + "\chunks2.vpp_pc"
	destpath = sr2path
	AddMessage("Copying chunks2.vpp_pc...")
	copy_file(sourcepath, destpath, fbArchive)

	Prog_Val+=1
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)

	sourcepath = _REBUILD_VPP_PC + "\chunks3.vpp_pc"
	destpath = sr2path
	AddMessage("Copying chunks3.vpp_pc...")
	copy_file(sourcepath, destpath, fbArchive)
	
	Prog_Val+=1
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
	
	sourcepath = _REBUILD_VPP_PC + "\pegs.vpp_pc"
	destpath = sr2path
	AddMessage("Copying pegs.vpp_pc...")
	copy_file(sourcepath, destpath, fbArchive)

	sleep 2000
	DelMessage("Copying chunks1.vpp_pc...")
	sleep 1000
	DelMessage("Copying chunks2.vpp_pc...")
	sleep 1000
	DelMessage("Copying chunks3.vpp_pc...")
	sleep 1000
	DelMessage("Copying pegs.vpp_pc...")

	Prog_Val+=1
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
	
	sleep 1000
	
	Prog_Val = 0
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
	
	MessageBox(_hDlg, "The new file was copied successfully.", @"Info", MB_OK or MB_APPLMODAL)
	
	THREAD_STATUS = THREAD_NOT_IN_USE
	THREAD_RETURN = F_NOERROR
end sub

'----------------------------------------------------------
'	delete all chunks data and initialize
'----------------------------------------------------------
sub init_chunks_data()
	dim as s32 i, j
	dim as string path, fullpath
	dim as zstring*256 ztemp
	
	THREAD_STATUS = THREAD_IN_USE

	AddMessage("Delete chunks texture data...")

	'	Progress bar
	Prog_Max = total_chunks
	Prog_Val = 0
	SendMessage(hProg, PBM_SETRANGE, 0, MAKELPARAM( 0, Prog_Max ))
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
	display_time()

	for i = 0 to 2
		select case i
			case 0
				path = "chunks1\"
			case 1
				path = "chunks2\"
			case 2
				path = "chunks3\"				
		end select
		for j = 0 to chunks(i).num - 1
		
			fullpath = path + chunks(i).chunk(j).name
			AddMessage(chunks(i).chunk(j).name)
			delete_file(fullpath + "\*.*", fbArchive)
			RmDir(fullpath)
		
			DelMessage("")
			Prog_Val+=1
			SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
			display_time()

		next
	next

	Prog_Val = 0
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
	DelMessage("")

	THREAD_STATUS = THREAD_NOT_IN_USE
	THREAD_RETURN = 0
	THREAD_ERROR = F_NOERROR
	
end sub

'----------------------------------------------------------
'	compare png image (same image return true)
'----------------------------------------------------------
function compare_image(srcfile as string, dstfile as string) as double
	dim as s32 src_w, src_h, dst_w, dst_h, bypp, diff_flag
	dim as u32 n, size
	dim as double psnr

	dim srcimage(any) as u8
	dim dstimage(any) as u8
	
	libpng.load_png(srcfile, srcimage(), src_w, src_h, bypp)
	libpng.load_png(dstfile, dstimage(), dst_w, dst_h, bypp)

	if (src_w <> dst_w) then return(false)
	if (src_h <> dst_h) then return(false)

	size = src_w * src_h * 4

	diff_flag = true
	for n = 0 to size - 1
		if (srcimage(n) <> dstimage(n)) then
			diff_flag = false
			exit for
		endif
	next
	return(diff_flag)
end function

'----------------------------------------------------------
'	delete file (returns success:false fail:true)
'----------------------------------------------------------
function delete_file(destpath as string, attrib_mask as s32) as s32
	dim as string path, fname, dfname
	
	path = libfio.get_directory(destpath)
	fname = dir(destpath, attrib_mask)	
	do while (fname <> "")
		dfname = path + fname
		fname = dir()
		if (kill(dfname)) then return(true)
	loop
	return(false)
end function

'----------------------------------------------------------
'	copy file (returns success:false fail:true)
'----------------------------------------------------------
function copy_file(sourcepath as string, destpath as string, attrib_mask as s32) as s32
	dim as string spath, fname, sfname, dfname
	
	spath = libfio.get_directory(sourcepath)
	fname = dir(sourcepath, attrib_mask)
	do while (fname <> "")
		sfname = spath + fname
		dfname = destpath + "\" + fname
		fname = dir()
		if (filecopy(sfname, dfname)) then return(true)
	loop
	return(false)
end function

'----------------------------------------------------------
'	move one file (returns success:false fail:true)
'----------------------------------------------------------
function move_onefile(sourcepath as string, destpath as string) as s32
	if (filecopy(sourcepath, destpath)) then
		return(true)
	else
		if (kill(sourcepath)) then return(true)
	endif
	return(false)
end function

'----------------------------------------------------------
'	check drive existence
'----------------------------------------------------------
function drive_exist(path as string) as s32
	dim as s32 attrib_mask
	dim as string drivepath, retstr
	
	attrib_mask = fbDirectory or fbArchive or fbHidden or fbSystem
	drivepath = left(path, 2) + "\NUL"
	
	retstr = dir(drivepath, attrib_mask)
	if (retstr =  "NUL") then
		return(F_NOERROR)
	else
		return(F_ERROR)
	endif
end function

'----------------------------------------------------------
'	prepare temporary folders
'----------------------------------------------------------
function init_temporary_folder() as s32
	dim as s32 ret
	dim as string destpath
	dim as zstring*256 ztemp

	if (temppath = "") then
		ztemp = !"temporary folder path is not set.\r\nPlease set temporary folder path."
		MessageBox(NULL, ztemp, @"Error", MB_OK)
		return(F_ERROR)
	else
		'	check drive existence
		if (mid(temppath, 2, 1) = ":") then
			if (drive_exist(temppath) = F_ERROR) then
				ztemp = "temporary folder drive " + left(temppath, 2) + " not exist."
				MessageBox(NULL, ztemp, @"Error", MB_OK)
				return(F_ERROR)
			endif
		endif
		'	preparation for temporary folder
		if (dir(temppath, fbDirectory) = "") then	'	folder not exist.
			'	make temporary folder
			ret = mkdir(temppath)
			if (ret) then
				'	mkdir fail
				ztemp = "Error:can't make temporary folder " + temppath
				MessageBox(NULL, ztemp, @"Error", MB_OK)
				return(F_ERROR)
			endif
		endif
		destpath = temppath + _TEMP_CHUNKS
		if (dir(destpath, fbDirectory) = "") then mkdir(destpath)
		destpath = temppath + _TEMP_DDS
		if (dir(destpath, fbDirectory) = "") then mkdir(destpath)
		destpath = temppath + _TEMP_PEGS
		if (dir(destpath, fbDirectory) = "") then mkdir(destpath)
	endif
	return(F_NOERROR)	'	no error
end function

'--------------------------------------------------------------------
'	get image colour as string
'--------------------------------------------------------------------
function get_image_colour_string(n as s32) as string
	select case n
		case DDS_sRGB_COLOUR
			return("sRGB")
		case DDS_LINEAR_COLOUR
			return("LINEAR")
		case else
			return("Undefined")
	end select
end function

'----------------------------------------------------------
'	display settings list (0:path settings, 1:dds settings
'----------------------------------------------------------
function get_dds_settings() as string
	dim as string filter, image_colour, stemp

	filter = dds.dds_director.mipmap(DDS_MIPMAP_DXT1).resize_filter
	image_colour = get_image_colour_string(dds.dds_director.mipmap(DDS_MIPMAP_DXT1).colourspace)
	stemp += "DXT1, " + filter + ", " + image_colour + !"\r\n"

	filter = dds.dds_director.mipmap(DDS_MIPMAP_DXT3_DXT5).resize_filter
	image_colour = get_image_colour_string(dds.dds_director.mipmap(DDS_MIPMAP_DXT3_DXT5).colourspace)
	stemp += "DXT3/5, " + filter + ", " + image_colour + !"\r\n"

	filter = dds.dds_director.mipmap(DDS_MIPMAP_NORMALMAP).resize_filter
	image_colour = get_image_colour_string(dds.dds_director.mipmap(DDS_MIPMAP_NORMALMAP).colourspace)
	stemp += "Normalmap, " + filter + ", " + image_colour + !"\r\n"

	return(stemp)
end function

'----------------------------------------------------------
'	get textures in trash folder and return total number
'----------------------------------------------------------
function get_trash_textures() as s32
	dim as s32 attrib_mask, n
	dim as string path, retstr
	
	path = "new_textures\trash\*.png"
	
	attrib_mask = fbArchive
	n = 0
	retstr = dir(path, attrib_mask)
	if (retstr <> "") then
		textures(n).name = retstr
		textures(n).replaceflag = false
		n+=1
	else
		return(n)
	endif

	do 
		retstr = dir()
		if (retstr <> "") then
			textures(n).name = retstr
			textures(n).replaceflag = false
			n+=1
		else
			exit do
		endif
	loop while (TRUE)
	
	return(n)
	
end function

'----------------------------------------------------------
'	get new textures name and return total number 
'----------------------------------------------------------
function get_new_textures() as s32
	dim as s32 attrib_mask, n
	dim as string path, retstr
	
	path = "new_textures\*.png"
	
	attrib_mask = fbArchive
	n = 0
	retstr = dir(path, attrib_mask)
	if (retstr <> "") then
		textures(n).name = retstr
		textures(n).replaceflag = false
		n+=1
	else
		return(n)
	endif

	do 
		retstr = dir()
		if (retstr <> "") then
			textures(n).name = retstr
			textures(n).replaceflag = false
			n+=1
		else
			exit do
		endif
	loop while (TRUE)
	
	return(n)

end function

'----------------------------------------------------------
'	get chunk names from list file
'	return total chunks number
'----------------------------------------------------------
function get_chunk_names_from_file(sourcefile as string) as s32
	dim as s32 ret, cn, n = 0, total = 0
	dim as string ename, value
	dim as zstring*256 ztemp
	
	ret = xml.read_doc(sourcefile)
	if (ret <> XML_NO_ERROR) then
		cprint using "ERROR:Can't open &"; sourcefile
		MessageBox(NULL, "Can't open " + sourcefile, @"Error", MB_OK)
		return(0)
	endif
	xml.read_element(ename, value)	' <ChunkNameList>
	if (ename <> "ChunkNameList") then
		ztemp = sourcefile + " is not chunknamelist file"
		MessageBox(NULL, ztemp, @"Error", MB_OK)
		xml.close_doc()
		return(0)
	endif
	xml.read_element(ename, value)	' <Entries>
	do
		' read 1 entry
		xml.read_element(ename, value)	' <Entry>
		select case ename
			case "Entry"
				xml.read_element(ename, value)	' <Name>chunks*</Name>
				n = 0
				select case value
					case "chunks1"
						cn = 0
					case "chunks2"
						cn = 1
					case "chunks3"
						cn = 2
					case else
				end select
				xml.read_element(ename, value)	' <Chunks>
				do
					xml.read_element(ename, value)	' <Chunk>***</Chunk> or </Entry>
					select case ename
						case "Chunk"
							chunks(cn).chunk(n).name = value
							chunks(cn).chunk(n).replaceflag = false
							n+=1
							if (n > MAX_CHUNKS) then
								cprint using "ERROR:Some syntax error found in &"; sourcefile
								MessageBox(NULL, "Some syntax error found in " + sourcefile, @"Error", MB_OK)
								xml.close_doc()
								return(0)
							endif
						case "/Chunks"
							chunks(cn).num = n
							total+=n
						case "/Entry"
							exit do
						case else
					end select
				loop
			case "/Entries"
				exit do
			case else
		end select
	loop

	xml.close_doc()

	return(total)

end function

'----------------------------------------------------------
'	get chunk names in chunks1\, chunks2\, chunks3\
'	return total chunks number 
'----------------------------------------------------------
function get_chunk_names() as s32
	dim as s32 i, n, total
	dim as string path, retstr

	total = 0
	for i = 0 to 2

		select case i
			case 0
				path = "chunks1\sr2*"
			case 1
				path = "chunks2\sr2*"
			case 2
				path = "chunks3\sr2*"
			case else
		end select
	
		n = 0
		retstr = dir(path, fbDirectory)
		if (retstr <> "") then		' there's a file
			chunks(i).chunk(n).name = libfio.get_filename_only(retstr)
			chunks(i).chunk(n).replaceflag = false
			n+=1
			do 
				retstr = dir()		' dir() returns next filename
				if (retstr <> "") then
					chunks(i).chunk(n).name = libfio.get_filename_only(retstr)
					chunks(i).chunk(n).replaceflag = false
					n+=1
				else
					chunks(i).num = n	' total number in chunks(i)
					exit do
				endif
			loop while (1)
		else
			' there's no file
			chunks(i).num = n		' total number in chunks(i) is 0	(error)
		endif
		total += n
	next
	
	return(total)

end function

'----------------------------------------------------------
'	restore original texture
'----------------------------------------------------------
function restore_original_textures(texn as s32) as s32
	dim as s32 attrib_mask, cn, i, j, k, n, ctexn, cx, cy, ret
	dim as s32 processed_chunks, progress_percent
	dim as string filespec, path, chunkpath, texpath, ddstexpath, sourcepath, destpath
	dim as string sourcename, destname, argstring, retstr
	
	attrib_mask = fbArchive

	for cn = 0 to 2

		select case cn
			case 0
				path = "chunks1\"
			case 1
				path = "chunks2\"
			case 2
				path = "chunks3\"
			case else			
		end select
	
		'	replace texture images
		for i = 0 to chunks(cn).num - 1

			'	search textures and revert to original
			for j = 0 to texn - 1
				texpath = path + chunks(cn).chunk(i).name + "\" + textures(j).name
				ddstexpath = libfio.change_filename_ext(texpath, ".dds")
				if (fileexists(texpath)) then	' same texture found
					sourcepath = libfio.change_filename_ext(texpath, "(org).png")
					destpath = texpath
					if (fileexists(sourcepath)) then	'	check (org).png file existence
						'	delete replaced png and dds
						kill(texpath)
						kill(ddstexpath)
						'	revert to original
						sourcepath = libfio.change_filename_ext(texpath, "(org).png")
						destpath = texpath
						name(sourcepath, texpath)	'	<texturename>(org).png -> texturename.png
						sourcepath = libfio.change_filename_ext(texpath, "(org).dds")
						destpath = ddstexpath
						name(sourcepath, destpath)	'	<texturename>(org).dds -> texturename.dds
						'	set rebuild flag
						chunks(cn).chunk(i).replaceflag = true
					endif
				endif
			next j
		next i
	next cn

	for i = 0 to texn - 1
		sourcepath = "new_textures\trash\" + textures(i).name
		destpath = "new_textures\trash\recycled\" + textures(i).name
		ret = move_onefile(sourcepath, destpath)
		if (ret) then return(F_ERROR)
	next i

	return(F_NOERROR)
	
end function

'----------------------------------------------------------
'	replace texture in each chunk and rebuild g_peg_pc/peg_pc
'----------------------------------------------------------
function replace_textures(cn as s32, texn as s32) as s32
	static as s32 processed_chunks = 0
	dim as s32 attrib_mask, i, j, k, n, ctexn, orgtexn, cx, cy, ret
	dim as string filespec, path, chunkpath, texpath, ddstexpath
	dim as string chunktemppath, pegtemppath, ddstemppath, renamepath, newtexpath, sourcepath, destpath
	dim as string sourcename, destname, argstring, retstr
	dim as string currtex(800)
	dim as zstring*256 ztemp

	attrib_mask = fbArchive
	
	select case cn
		case 0
			path = "chunks1\"
		case 1
			path = "chunks2\"
		case 2
			path = "chunks3\"
		case else			
	end select

	pegtemppath = temppath + _TEMP_CHUNKS + "\"
	ddstemppath = temppath + _TEMP_DDS + "\"

	'	replace texture images
	for i = 0 to chunks(cn).num - 1
		'               012345678901234567890123456789
		'cprint using "> \                            \ replacing textures..."; chunks(cn).chunk(i).name;
		'ztemp = libstr.strform_str("Left", 30, chunks(cn).chunk(i).name) + " replacing textures..."
		ztemp = libstr.strform_str("Left", 30, chunks(cn).chunk(i).name) + " replacing textures..."
		AddMessage(ztemp)

		chunkpath = path + chunks(cn).chunk(i).name + "\"
		filespec = chunkpath + "*.png"

		'	get existence texture names in the chunk
		ctexn = 0
		retstr = dir(filespec, attrib_mask)
		do while retstr <> ""
			currtex(ctexn) = retstr
			ctexn+=1
			retstr = dir()
		loop

		'	search textures and replace to new one
		for j = 0 to ctexn - 1
			for k = 0 to texn - 1
				if (currtex(j) = textures(k).name) then		'	same name file found
					chunks(cn).chunk(i).replaceflag = true	'	set replaceflag
					texpath = path + chunks(cn).chunk(i).name + "\" + textures(k).name
					newtexpath = "new_textures\" + textures(k).name

					ret = compare_image(texpath, newtexpath)
					if (not(ret)) then 								'	new texture is different from existing one

						ddstexpath = libfio.change_filename_ext(texpath, ".dds")

						'	backup original texture as texturename(org).png when texturename(org).png is not exist.
						renamepath = libfio.change_filename_ext(texpath, "(org).png")
						if (not(fileexists(renamepath))) then
							name(texpath, renamepath)
						end if

						'	backup original texture as texturename(org).dds when texturename(org).dds is not exist.
						renamepath = libfio.change_filename_ext(texpath, "(org).dds")
						if (not(fileexists(renamepath))) then
							name(ddstexpath, renamepath)
						else
							'	delete dds file
							kill(ddstexpath)
						end if
	
						'	replace png texture image file
						filecopy(newtexpath, texpath)
						
						'	set replace flag 
						if (textures(k).replaceflag = false) then
							'	set flag
							textures(k).replaceflag = true
							'	store replaced texture index
							replaced(replacedn).index = k
							replacedn+=1
						else
							'	copy already converted dds file
							sourcepath = ddstemppath + textures(k).name
							sourcepath = libfio.change_filename_ext(sourcepath, ".dds")
							destpath = libfio.change_filename_ext(texpath, ".dds")
							filecopy(sourcepath, destpath)
						endif
					else
						'	new texture is same to original one
						if (textures(k).replaceflag = false) then	'	the new texture not converted yet
							ddstexpath = libfio.change_filename_ext(texpath, ".dds")
							'	backup original texture as texturename(org).png when texturename(org).png is not exist.
							renamepath = libfio.change_filename_ext(texpath, "(org).png")
							if (not(fileexists(renamepath))) then
								name(texpath, renamepath)
								filecopy(newtexpath, texpath)
							end if
							'	backup original texture as texturename(org).dds when texturename(org).dds is not exist.
							renamepath = libfio.change_filename_ext(texpath, "(org).dds")
							if (not(fileexists(renamepath))) then
								name(ddstexpath, renamepath)
							else
								'	delete dds file
								kill(ddstexpath)
							end if
							textures(k).replaceflag = true
							'	store replaced texture index
							replaced(replacedn).index = k
							replacedn+=1
						else
							'	copy already converted dds file
							sourcepath = ddstemppath + textures(k).name
							sourcepath = libfio.change_filename_ext(sourcepath, ".dds")
							destpath = libfio.change_filename_ext(texpath, ".dds")
							ret = filecopy(sourcepath, destpath)
							if (ret) then
								'cprint using "can't copy file &."; textures(k).name + ".dds"
								return(true)
							endif
						endif
					endif
				endif
			next k
		next j

		DelMessage(ztemp)

		if (chunks(cn).chunk(i).replaceflag = true) then
			

			'	get *(org).png number
			filespec = chunkpath + "*(org).png"
			orgtexn = 0
			retstr = dir(filespec, attrib_mask)
			do while retstr <> ""
				orgtexn+=1
				retstr = dir()
			loop		

			'	get peg_desc filename to rebuild peg_pc and gpeg_pc
			sourcename = path + chunks(cn).chunk(i).name + "\" + chunks(cn).chunk(i).name + ".peg_desc"
			
			'	rebuild peg_pc and gpeg_pc
			'               012345678901234567890123456789
			'cprint using "> \                            \ rebuilding... entrycount:####"; chunks(cn).chunk(i).name; ctexn - orgtexn;
			ztemp = libstr.strform_str("Left", 30, chunks(cn).chunk(i).name) + " rebuilding... entrycount:" + libstr.strform_val(" ", 4, 0, ctexn - orgtexn)
			AddMessage(ztemp)
			n = peg_pc.rebuild_peg(sourcename, pegtemppath)
			if (n = F_ERROR) then
				ztemp = "Fail to rebuild " + chunks(cn).chunk(i).name
				MessageBox(NULL, ztemp, @"Error", MB_OK)
				return(F_ERROR)
			endif
			DelMessage("")						

			'	move .peg_pc file
			sourcepath = pegtemppath + chunks(cn).chunk(i).name + ".peg_pc"
			destpath = temppath + _TEMP_PEGS + "\" + chunks(cn).chunk(i).name + ".peg_pc"
			ret = move_onefile(sourcepath, destpath)
			if (ret) then
				ztemp = "can't move file " + chunks(cn).chunk(i).name + ".peg_pc"
				MessageBox(NULL, ztemp, @"Error", MB_OK)
				return(F_ERROR)
			endif
			
			'	copy new dds file to temporary folder
			for j = processed_textures to replacedn - 1
				sourcepath = path + chunks(cn).chunk(i).name + "\" + textures(replaced(j).index).name
				sourcepath = libfio.change_filename_ext(sourcepath, ".dds")
				destpath = ddstemppath + textures(replaced(j).index).name
				destpath = libfio.change_filename_ext(destpath, ".dds")
				filecopy(sourcepath, destpath)
			next
			processed_textures = replacedn
		endif

		'	progress bar
		Prog_Val +=1
		SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
		display_time()

		if (THREAD_ABORT = TRUE) then
			'abort_thread()
			return(F_ERROR)
		endif

	next i
	
	return(F_NOERROR)
	
end function

'----------------------------------------------------------
'	replace texture and rebuild vpp_pc
'----------------------------------------------------------
sub replace_textures_and_rebuild()
	dim as s32 i, n, texn, ret
	dim as string kstr, sourcepath, destpath, cmdstring, resize_method
	dim as zstring*256 ztemp

	THREAD_STATUS = IDM_MAKE_REBUILD
	THREAD_ABORT = FALSE

	ztemp = "Replace textures and rebuid vpp_pc files."
	AddMessage(ztemp)

	delete_temporary_files()

	'	processing texture replacement
	ztemp = "Gathering file infomation..."
	AddMessage(ztemp)
	
	'	get chunk names and total number
	total_chunks = get_chunk_names()
	
	'------------------------------------------------------
	'	restore original texture when trash folder is not empty
	'------------------------------------------------------
	' get textures in trash folder
	texn = get_trash_textures()
	DelMessage("")
	
	if (texn < 1) then
		'	nop
	else
		ztemp = "Restoring original textures..."
		AddMessage(ztemp)
		'	restore original textures.
		ret = restore_original_textures(texn)
		if (ret) then
			ztemp = "restoring original textures failed"
			MessageBox(NULL, ztemp, @"Error", MB_OK)
			THREAD_RETURN = ret
			THREAD_STATUS = THREAD_NOT_IN_USE
			return
		else
			DelMessage("")
		endif
	endif

	'	init progressbar
	Prog_Val = 0
	Prog_Max = total_chunks + 3
	SendMessage(hProg, PBM_SETRANGE, 0, MAKELPARAM(0, Prog_Max))
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
	display_time()

	'	get new textures
	texn = get_new_textures()
	
	'		extract original peg files to temp folder
	if (THREAD_ABORT = TRUE) then
		abort_thread()
		return
	endif
	ztemp = "Extracting pegs.vpp_pc files to temporary folder..."
	AddMessage(ztemp)

	sourcepath = _ORIGINAL_VPP_PC + "\pegs.vpp_pc"
	destpath = temppath + _TEMP_PEGS
	SaintsRow2_FileFormats.UnpackFile(sourcepath, destpath)

	DelMessage("")
	
'	progress_bar
	Prog_Val+=1
	SendMessage(hProg, PBM_SETPOS, Prog_val, 0)
	display_time()
	
	' ------------------------------------
	' 	chunks
	' ------------------------------------
	'	replace chunks1, 2, 3 textures and rebuild g_peg_pc and peg_pc.
	for i = 0 to 2
	
		'	extract original chunks data to temporary folder
		if (THREAD_ABORT = TRUE) then
			abort_thread()
			return
		endif

		ztemp = "Extracting chunks" + libstr.strform_val(" ", 1, 0, i + 1) + ".vpp__pc file to temporary folder..."
		AddMessage(ztemp)
		select case i
			case 0
				sourcepath = _ORIGINAL_VPP_PC + "\chunks1.vpp_pc"
			case 1
				sourcepath = _ORIGINAL_VPP_PC + "\chunks2.vpp_pc"
			case 2
				sourcepath = _ORIGINAL_VPP_PC + "\chunks3.vpp_pc"
			case else
		end select
		
		destpath = temppath + _TEMP_CHUNKS
		SaintsRow2_Fileformats.UnpackFile(sourcepath, destpath)
		DelMessage("")
		
		'	replace textures and rebuid new gpeg_pc and peg_pc
		ret = replace_textures(i, texn)

		if (ret = F_ERROR) then
			abort_thread()
			return
		endif

		'	chunks(i) g_peg_pc
		'cprint using "> building chunks#.vpp__pc..."; i + 1;

		if (THREAD_ABORT = TRUE) then
			abort_thread()
			return
		endif

		ztemp = "Building chunks" + libstr.strform_val(" ", 1, 0, i + 1) + ".vpp__pc..."
		AddMessage(ztemp)	
		'		re-package chunks.vpp_pc
		sourcepath = temppath + _TEMP_CHUNKS
		select case i
			case 0
				destpath = _REBUILD_VPP_PC + "\chunks1.vpp_pc"
			case 1
				destpath = _REBUILD_VPP_PC + "\chunks2.vpp_pc"
			case 2
				destpath = _REBUILD_VPP_PC + "\chunks3.vpp_pc"
			case else
		end select

		SaintsRow2_FileFormats.PackFile(sourcepath, destpath)

		DelMessage("")

		'		cleanup temporary chunks folder
		destpath = temppath + _TEMP_CHUNKS + "\*.*"
		delete_file(destpath, FbArchive)
	next

	' ------------------------------------
	' 	pegs
	' ------------------------------------

	if (THREAD_ABORT = TRUE) then
		abort_thread()
		return
	endif

	ztemp = "Building pegs.vpp_pc..."
	AddMessage(ztemp)
	
	'		re-package pegs.vpp_pc
	sourcepath = temppath + _TEMP_PEGS
	destpath = _REBUILD_VPP_PC + "\pegs.vpp_pc"
	SaintsRow2_FileFormats.PackFile(sourcepath, destpath)
	DelMessage("")

	Prog_Val+=1
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
	display_time()

	'		cleanup peg temporary files
	ztemp = "Delete temporary files..."
	AddMessage(ztemp)

	destpath = temppath + _TEMP_PEGS + "\*.*"
	delete_file(destpath, FbArchive)

	'	cleanup temporary dds folder
	destpath = temppath + _TEMP_DDS + "\*.*"
	delete_file(destpath, FbArchive)

	DelMessage("")

	'	progress 100%
	Prog_Val+=1
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
	display_time()

	DelMessage("")
	ztemp = "All processes done."
	AddMessage(ztemp)

	sleep 1000
	
	DelMessage("*")

	Prog_Val = 0
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)

	THREAD_STATUS = THREAD_NOT_IN_USE
	THREAD_ABORT = FALSE

	'	Make-Abort Grayed out
	EnableMenuItem(hMenu, IDM_MAKE_ABORT, MF_GRAYED)
	DrawMenuBar(hMenu)

	return
	
end sub

'----------------------------------------------------------
'	make chunks texture data
'----------------------------------------------------------
sub make_chunks_texture_data()
	dim as s32 i, j, n, texn, ret
	dim as s32 ln, ln2
	dim as string sourcepath, destpath, cmdstring, resize_method
	dim as zstring*256 ztemp

	THREAD_STATUS = IDM_MAKE_EXTRACT

	ztemp = "Extract chunks1, chunks2, chunks3 texture data."	'	first line
	AddMessage(ztemp)

	delete_temporary_files()

	'------------------------------------------------------
	'	create chunks1, chunks2, chunks3 data
	'------------------------------------------------------
	'		extract original peg files to temp folder

	sourcepath = "chunkname_list.dat"
	total_chunks = get_chunk_names_from_file(sourcepath)
	
	'	Progress bar
	Prog_Max = total_chunks + 1 + 3
	Prog_Val = 0
	SendMessage(hProg, PBM_SETRANGE, 0, MAKELPARAM( 0, Prog_Max ))
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
	display_time()

	ztemp = "Extracting pegs.vpp_pc to temporary folder..."
	AddMessage(ztemp)

	sourcepath = _ORIGINAL_VPP_PC + "\pegs.vpp_pc"
	destpath = temppath + _TEMP_PEGS
	SaintsRow2_FileFormats.UnpackFile(sourcepath, destpath)

	DelMessage("")

	Prog_Val+=1
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
	display_time()

	for i = 0 to 2			
		'	extract original chunks data to temporary folder
		ztemp = "Extracting chunks" + libstr.strform_val("", 1, 0, i+1) + ".vpp_pc to temporary folder..."
		AddMessage(ztemp)

		select case i
			case 0
				sourcepath = _ORIGINAL_VPP_PC + "\chunks1.vpp_pc"
			case 1
				sourcepath = _ORIGINAL_VPP_PC + "\chunks2.vpp_pc"
			case 2
				sourcepath = _ORIGINAL_VPP_PC + "\chunks3.vpp_pc"
			case else
		end select
		
		destpath = temppath + _TEMP_CHUNKS
		SaintsRow2_FileFormats.UnpackFile(sourcepath, destpath)

		DelMessage("")
		
		'	progress bar
		Prog_Val+=1
		SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
		display_time()

		if (THREAD_ABORT = TRUE) then
			abort_thread()
			return
		endif
		
		'	copy peg_pc to _TEMP_CHUNKS
		for j = 0 to chunks(i).num - 1
			sourcepath = temppath + _TEMP_PEGS + "\" + chunks(i).chunk(j).name + ".peg_pc"
			destpath = temppath + _TEMP_CHUNKS
			copy_file(sourcepath, destpath, fbArchive)
		next

		if (THREAD_ABORT = TRUE) then
			abort_thread()
			return
		endif
				
		'	extract gpeg_pc files
		for j = 0 to chunks(i).num - 1

			if (THREAD_ABORT = TRUE) then
				abort_thread()
				return
			endif

			sourcepath = temppath + _TEMP_CHUNKS + "\" + chunks(i).chunk(j).name + ".peg_pc"
			destpath = "chunks" + str(i+1) + "\" + chunks(i).chunk(j).name
			'	check folder existence
			if (dir(destpath, fbDirectory) <> "") then
				'	cleanup files
				ztemp = chunks(i).chunk(j).name + " delete existing files..."
				AddMessage(ztemp)

				delete_file(destpath + "\*.*", fbArchive)

				DelMessage("")
			endif

			'	extracting
			ztemp = chunks(i).chunk(j).name + " extract texture data..."
			AddMessage(ztemp)
			peg_pc.extract_peg(sourcepath, destpath)
			DelMessage("")
			
			'	progress bar
			Prog_Val+=1
			SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
			display_time()
		
		next
		'		cleanup chunks temporary files
		destpath = temppath + _TEMP_CHUNKS + "\*.*"
		delete_file(destpath, FbArchive)
	next

	'		cleanup peg temporary files
	destpath = temppath + _TEMP_PEGS + "\*.*"
	delete_file(destpath, FbArchive)	

	'	reset progress bar
	Prog_Val = 0
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
	ztemp = "All processes done."
	AddMessage(ztemp)
	sleep 1000
	'	clear messages
	DelMessage("")
	DelMessage("")

	THREAD_STATUS = THREAD_NOT_IN_USE
	THREAD_ABORT = FALSE

end sub

'----------------------------------------------------------
'	make chunks texture statistics
'----------------------------------------------------------
sub make_texture_statistics()
	dim chunk_textures(MAX_TEXTURES) as chunk_textures_data
	dim as s32 cn, i, j, k, texn, fh
	dim as s32 mip, imgw, imgh, pixform, flag, exitflag
	dim as string retstr, path, chunkpath, filespec, fullpath
	dim as zstring*64 ztemp

	'	Thread state
	THREAD_STATUS = THREAD_IN_USE
	
	'	Progress bar
	Prog_Max = 4
	Prog_Val = 0
	SendMessage(hProg, PBM_SETRANGE, 0, MAKELPARAM( 0, Prog_Max ))
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)

	AddMessage("Making chunks texture statistics...")
	
	exitflag = false
	texn = 0
	for cn = 0 to 2

		select case cn
			case 0
				path = "chunks1\"
			case 1
				path = "chunks2\"
			case 2
				path = "chunks3\"
			case else
		end select

		for i = 0 to chunks(cn).num - 1

			ztemp = "searching " + chunks(cn).chunk(i).name + "..."
			AddMessage(ztemp)
						
			chunkpath = path + chunks(cn).chunk(i).name + "\"	
			filespec = chunkpath + "*.dds"

			'	get existence texture names in the chunk
			retstr = dir(filespec, fbArchive)
			if (retstr <> "") then
				flag = FALSE
				if (texn > 0) then
					k = texn - 1
					for j = 0 to k
						if (retstr = chunk_textures(j).name) then
							' this texture is already exists
							chunk_textures(j).num += 1
							flag = TRUE
							exit for
						endif
					next
					if (flag = FALSE) then
						' new texture found
						chunk_textures(texn).name = retstr
						chunk_textures(texn).chunkname = chunks(cn).chunk(i).name
						fullpath = chunkpath + retstr
						mip = dds.load_ddsheader(fullpath, imgw, imgh, pixform)
						chunk_textures(texn).width = imgw
						chunk_textures(texn).height = imgh
						chunk_textures(texn).format = pixform
						chunk_textures(texn).num = 1
						texn+=1
						if (texn > MAX_TEXTURES) then
							exitflag = TRUE
							exit for
						endif
					endif
				else
					' first texture found
					chunk_textures(texn).name = retstr
					chunk_textures(texn).chunkname = chunks(cn).chunk(i).name
					fullpath = chunkpath + retstr
					mip = dds.load_ddsheader(fullpath, imgw, imgh, pixform)
					chunk_textures(texn).width = imgw
					chunk_textures(texn).height = imgh
					chunk_textures(texn).format = pixform
					chunk_textures(texn).num = 1
					texn+=1
				endif
				
				do
					retstr = dir()
					if (retstr <> "") then
						flag = FALSE
						k = texn - 1
						for j = 0 to k
							if (retstr = chunk_textures(j).name) then
								' this texture is already exists
								chunk_textures(j).num += 1
								flag = TRUE
								exit for
							endif
						next
						if (flag = FALSE) then
							' new texture found
							chunk_textures(texn).name = retstr
							chunk_textures(texn).chunkname = chunks(cn).chunk(i).name
							fullpath = chunkpath + retstr
							mip = dds.load_ddsheader(fullpath, imgw, imgh, pixform)
							chunk_textures(texn).width = imgw
							chunk_textures(texn).height = imgh
							chunk_textures(texn).format = pixform
							chunk_textures(texn).num = 1
							texn+=1
							if (texn > MAX_TEXTURES) then
								exitflag = TRUE
								exit do
							endif
						endif
					else
						' no more file
						exit do
					endif
				loop while(TRUE)
				if (exitflag = TRUE) then
					exit for
				endif
			endif
			DelMessage("")
		next i
		if (exitflag = TRUE) then 
			DelMessage("")
			exit for
		endif
		Prog_Val+=1
		SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
	next cn

	if (exitflag) then
		Prog_Val = 0
		SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
		ztemp = "number of texture excess upper limit!"
		MessageBox(NULL, ztemp, @"Error", MB_OK)
		THREAD_STATUS = THREAD_NOT_IN_USE
		return
	endif
	
	'	sort texture data (simple bubble sort)
	AddMessage("sorting data...")

	do
		flag = FALSE
		for i = 1 to texn - 1
			if (chunk_textures(i - 1).num < chunk_textures(i).num) then
				swap chunk_textures(i - 1), chunk_textures(i) 
				flag = TRUE
			elseif (chunk_textures(i - 1).num = chunk_textures(i).num) then
				if (chunk_textures(i - 1).name > chunk_textures(i).name) then
					swap chunk_textures(i - 1), chunk_textures(i) 
					flag = TRUE					
				endif
			endif
		next
	loop while (flag = TRUE)

	DelMessage("")
	Prog_Val+=1
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)
		
	'	make report file
	
	fh = freefile
	open "texture_statistics.txt" for output as #fh
    print #fh, "Texture Name                             Num  Width  Height  Format   First Found"
    print #fh, "--------------------------------------+-----+------+------+----------+------------------------------"
'              "\                                    \| ####|  ####|  ####| \      \ | &"
'                                                                            A8R8G8B8    
	for i = 0 to texn - 1
		j = chunk_textures(i).num
		imgw = chunk_textures(i).width
		imgh = chunk_textures(i).height
		pixform = chunk_textures(i).format
		select case pixform
			case D3DFMT_DXT1
				retstr = "DXT1 "
			case D3DFMT_DXT3
				retstr = "DXT3 "
			case D3DFMT_DXT5
				retstr = "DXT5 "
			case D3DFMT_A8R8G8B8
				retstr = "A8R8G8B8"
			case D3DFMT_R5G6B5
				retstr = "R5G6B5"
			case else
				retstr = "?????"
		end select
		print #fh, using "\                                    \| ####|  ####|  ####| \      \ | &";chunk_textures(i).name; j; imgw; imgh; retstr; chunk_textures(i).chunkname
	next i
	
	close fh

	Prog_Val = 0
	SendMessage(hProg, PBM_SETPOS, Prog_Val, 0)

	DelMessage("")

	THREAD_STATUS = THREAD_NOT_IN_USE
	
end sub

'----------------------------------------------------------
'	write ini file
'----------------------------------------------------------
function write_inifile(filename as string) as s32
	dim as s32 ret, fh, i
	dim as string st
	
	ret = xml.new_doc(filename, "sr2texturetool")
	if (ret <> XML_NO_ERROR) then
		return(F_ERROR)
	endif
	
	xml.write_single_element("Version", PROGRAM_VERSION)
	xml.new_element("PathSettings")
		xml.write_single_element("SR2Path", sr2Path)
		xml.write_single_element("TempPath", temppath)
	xml.close_element()
	xml.new_element("PegDirector")
		xml.write_single_element("CheckImageSizeDiff", "FALSE")
		xml.write_single_element("Suppress_extracted", "TRUE")
	xml.close_element()
	xml.new_element("DDSDirector")
		xml.new_element("Const")
			xml.write_single_element("RenderTarget", "SR2")
			xml.write_single_element("SilentMode", "TRUE")
		xml.close_element()
		xml.new_element("DxtDirector")
			select case dds.dds_director.colourfit
				case squish_kColourClusterFit
					st = "ClusterFit"
				case squish_kColourRangeFit
					st = "RangeFit"
				case squish_kColourIterativeClusterFit
					st = "IterativeClusterFit"
				case else
					st = "ClusterFit"
			end select
			xml.write_single_element("ColourFit", st)
		xml.close_element()
		xml.new_element("MipmapDirectors")
			for i = 0 to 2
				xml.new_element("MipmapDirector")
				select case i
					case 0
						xml.write_single_element("Format", "DXT1")
					case 1
						xml.write_single_element("Format", "DXT3orDXT5")
					case 2
						xml.write_single_element("Format", "NORMALMAP")
					case else
				end select				
				st = dds.dds_director.mipmap(i).resize_filter
				xml.write_single_element("Filter", st)
				select case dds.dds_director.mipmap(i).colourspace
					case DDS_sRGB_COLOUR
						st = "sRGB"
					case DDS_LINEAR_COLOUR
						st = "LINEAR"
					case else
				end select
				xml.write_single_element("Colour", st)
				xml.close_element()
			next
		xml.close_element()	'	MipmapDirectors
	xml.close_element()	'	DDSDirector
	xml.close_doc()

	fh = freefile
	open filename for append as #fh
	
	print #fh, "//---------------------------------------------"
	print #fh, "//    Description Help"
	print #fh, "//---------------------------------------------"
	print #fh, "// <CheckImageSizeDiff>"
	print #fh, "//   it must always be FALSE in SR2ChunksTextureTool"
	print #fh, "// <Suppress_extracted>"
	print #fh, "//   FALSE(default)"
	print #fh, "//     add '_extracted' to extract folder name"
	print #fh, "//   TRUE"
	print #fh, "//     do not add '_extracted' to extract folder name"
	print #fh, "// <RenderTarget>"
	print #fh, "//   SR2     (generate mipmap to 4x4)"
	print #fh, "//   GENERIC (generate mipmap to 1x1)"
	print #fh, "// <ColourFit>"
	print #fh, "//   ClusterFit          - Use a slow but high quality colour compressor (default)"
	print #fh, "//   IterativeClusterFit - Use a very slow but very high quality colour compressor"
	print #fh, "//   RangeFit            - Use a fast but low quality colour compressor"
	print #fh, "// <Format> texture format"
	print #fh, "//   DXT1"
	print #fh, "//   DXT3orDXT5"
	print #fh, "//   NORMALMAP"
	print #fh, "// <Filter>  image resize filter"
	print #fh, "//   POINT LINEAR CUBIC FANT BOX TRIANGLE POINT_DITHER"
	print #fh, "//   LINEAR_DITHER CUBIC_DITHER FANT_DITHER BOX_DITHER TRIANGLE_DITHER"
	print #fh, "//   POINT_DITHER_DIFFUSION LINEAR_DITHER_DIFFUSION CUBIC_DITHER_DIFFUSION"
	print #fh, "//   FANT_DITHER_DIFFUSION BOX_DITHER_DIFFUSION TRIANGLE_DITHER_DIFFUSION"
	print #fh, "// <Colour> source image color"
	print #fh, "//   sRGB   for general texture"
	print #fh, "//   LINEAR for normalmap/tgn"
	close #fh
	
	return(F_NOERROR)

end function

'----------------------------------------------------------
'	read ini file
'----------------------------------------------------------
function read_inifile(filename as string) as s32
	dim as s32 i, n, ret, texture_format
	dim as string ename, evalue
	dim as zstring*256 ztemp
	
	'	open ini file
	ret = xml.read_doc(filename)
	if (ret <> XML_NO_ERROR) then
		ztemp = "Can't read " + _INIFILE_NAME
		MessageBox(NULL, ztemp, @"Error", MB_OK)
		xml.close_doc()
		return(F_ERROR)
	endif

	'	<sr2texturetool>
	ret = xml.read_element(ename, evalue)
	if (ename <> "sr2texturetool") then
		ztemp = _INIFILE_NAME + " <sr2texturetool> missing"
		MessageBox(NULL, ztemp, @"Error", MB_OK)
		xml.close_doc()
		return(F_ERROR)
	endif
	
	'	<Version>
	ret = xml.read_element(ename, evalue)
	if (ename <> "Version") then
		ztemp = _INIFILE_NAME + " missing. " + ename
		MessageBox(NULL, ztemp, @"Error", MB_OK)
		xml.close_doc()
		return(F_ERROR)
	elseif (val(evalue) <> val(PROGRAM_VERSION)) then
		ztemp = _INIFILE_NAME + " <Version> not match. " + evalue
		MessageBox(NULL, ztemp, @"Error", MB_OK)
		xml.close_doc()
		return(F_ERROR)
	endif

	'	<PathSettings>
	ret = xml.read_element(ename, evalue)
	if (ename <> "PathSettings") then
		ztemp = _INIFILE_NAME + " <PathSettings> missing"
		MessageBox(NULL, ztemp, @"Error", MB_OK)
		xml.close_doc()
		return(F_ERROR)
	endif
	do
		ret = xml.read_element(ename, evalue)
		select case ename
			case "SR2Path"
				sr2path = trim(evalue)
			case "TempPath"
				temppath = trim(evalue)
				if temppath = "" then
					ztemp = _INIFILE_NAME + " <TempPath> block is blank."
					MessageBox(NULL, ztemp, @"Error", MB_OK)
					xml.close_doc()
					return(F_ERROR)
				endif
			case "/PathSettings"
				exit do
			case else
				ztemp = _INIFILE_NAME + " <PathSettings> block syntax error. " + ename
				MessageBox(NULL, ztemp, @"Error", MB_OK)
				xml.close_doc()
				return(F_ERROR)
		end select
	loop while (true)

	'	<PegDirector>
	ret = xml.read_element(ename, evalue)
	if (ename <> "PegDirector") then
		ztemp = _INIFILE_NAME + " <PegDirector> missing. " + ename
		MessageBox(NULL, ztemp, @"Error", MB_OK)
		xml.close_doc()
		return(F_ERROR)
	endif
	
	do
		ret = xml.read_element(ename, evalue)
		select case ename
			case "CheckImageSizeDiff"
				select case lcase(evalue)
					case "true"
						peg_pc.peg_director.check_imagesize_difference = TRUE
					case "false"
						peg_pc.peg_director.check_imagesize_difference = FALSE
					case else						
						ztemp = _INIFILE_NAME + " <CheckImageSizeDiff> value wrong. " + evalue
						MessageBox(NULL, ztemp, @"Error", MB_OK)
						xml.close_doc()
						return(F_ERROR)
				end select
			case "Suppress_extracted"
				select case lcase(evalue)
					case "true"
						peg_pc.peg_director.suppress_extracted = TRUE
					case "false"
						peg_pc.peg_director.suppress_extracted = FALSE
					case else
						ztemp = _INIFILE_NAME + " <Suppress_extracted> value wrong. " + evalue
						MessageBox(NULL, ztemp, @"Error", MB_OK)
						xml.close_doc()
						return(F_ERROR)
				end select
			case "/PegDirector"
				exit do
			case else
				ztemp = _INIFILE_NAME + " <PegDirector> block syntax error. " + ename
				MessageBox(NULL, ztemp, @"Error", MB_OK)
				xml.close_doc()
				return(F_ERROR)
		end select
	loop while (true)

	'	<DDSDirector>
	ret = xml.read_element(ename, evalue)
	if (ename <> "DDSDirector") then
		ztemp = _INIFILE_NAME + " <DDSDirector> missing"
		MessageBox(NULL, ztemp, @"Error", MB_OK)
		xml.close_doc()
		return(F_ERROR)
	endif

	'	<Const>
	ret = xml.read_element(ename, evalue)	'Const
	if (ename <> "Const") then
		ztemp = _INIFILE_NAME + " <Const> missing"
		MessageBox(NULL, ztemp, @"Error", MB_OK)
		xml.close_doc()
		return(F_ERROR)
	endif
	do
		ret = xml.read_element(ename, evalue)
		select case ename
			case "RenderTarget"
				select case evalue
					case "SR2"
						dds.dds_director.render_target = DDS_SR2_ENGINE
					case "GENERIC"
						dds.dds_director.render_target = DDS_GENERIC_ENGINE
					case else
						ztemp = _INIFILE_NAME + " wrong description in <RenderTarget>. " + evalue
						MessageBox(NULL, ztemp, @"Error", MB_OK)
						xml.close_doc()
						return(F_ERROR)
				end select
			case "SilentMode"
				select case lcase(evalue)
					case "true"
						dds.dds_director.silentmode = TRUE
					case "false"
						dds.dds_director.silentmode = FALSE
					case else
						ztemp = _INIFILE_NAME + " wrong description in <SilentMode>. " + evalue
						MessageBox(NULL, ztemp, @"Error", MB_OK)
						xml.close_doc()
						return(F_ERROR)
				end select
			case "/Const"
				exit do
			case else
				ztemp = _INIFILE_NAME + " <Const> block syntax error. " + ename
				MessageBox(NULL, ztemp, @"Error", MB_OK)
				xml.close_doc()
				return(F_ERROR)
		end select
	loop while (true)

	'	<DxtDirector>
	ret = xml.read_element(ename, evalue)	'Const
	if (ename <> "DxtDirector") then
		ztemp = _INIFILE_NAME + " <DxtDirector> missing"
		MessageBox(NULL, ztemp, @"Error", MB_OK)
		xml.close_doc()
		return(F_ERROR)
	endif
	do
		ret = xml.read_element(ename, evalue)
		select case ename
			case "ColourFit"
				select case evalue
					case "ClusterFit"
						dds.dds_director.colourfit = squish.kColourClusterFit
					case "IterativeClusterFit"
						dds.dds_director.colourfit = squish.kColourIterativeClusterFit
					case "RangeFit"
						dds.dds_director.colourfit = squish.kColourRangeFit
					case else
						ztemp = _INIFILE_NAME + " wrong description in <ColourFit>. " + evalue
						MessageBox(NULL, ztemp, @"Error", MB_OK)
						xml.close_doc()
						return(F_ERROR)
				end select
			case "/DxtDirector"
				exit do
			case else
				ztemp = _INIFILE_NAME + " <DxtDirector> block syntax error. " + ename
				MessageBox(NULL, ztemp, @"Error", MB_OK)
				xml.close_doc()
				return(F_ERROR)
		end select
	loop while (true)

	'	read MipmapDirector settings
	ret = xml.read_element(ename, evalue)
	if (ename <> "MipmapDirectors") then
		ztemp = _INIFILE_NAME + " <MipmapDirectors> missing"
		MessageBox(NULL, ztemp, @"Error", MB_OK)
		xml.close_doc()
		return(F_ERROR)		
	endif
	do
		ret = xml.read_element(ename, evalue)
		select case ename
			case "MipmapDirector"
				'	nop
			case "Format"
				select case evalue
					case "NORMALMAP"
						texture_format = DDS_MIPMAP_NORMALMAP
					case "DXT1"
						texture_format = DDS_MIPMAP_DXT1
					case "DXT3orDXT5"
						texture_format = DDS_MIPMAP_DXT3_DXT5
					case else
						ztemp = _INIFILE_NAME + " wrong description in <Format>. " + evalue
						MessageBox(NULL, ztemp, @"Error", MB_OK)
						xml.close_doc()
						return(F_ERROR)
				end select
			case "Filter"
				dds.dds_director.mipmap(texture_format).resize_filter = trim(evalue)
			case "Colour"
				select case evalue
					case "sRGB"
						dds.dds_director.mipmap(texture_format).colourspace = DDS_sRGB_COLOUR
					case "LINEAR"
						dds.dds_director.mipmap(texture_format).colourspace = DDS_LINEAR_COLOUR
					case else
						ztemp = _INIFILE_NAME + " <Colour> value wrong. " + evalue
						MessageBox(NULL, ztemp, @"Error", MB_OK)
						xml.close_doc()
						return(F_ERROR)
				end select
			case "/MipmapDirector"
				'	nop
			case "/MipmapDirectors"
				exit do
			case else
				ztemp = _INIFILE_NAME + " <MipmapDirector> block syntax error. " + ename
				MessageBox(NULL, ztemp, @"Error", MB_OK)
				xml.close_doc()
				return(F_ERROR)
		end select
	loop while (true)

	ret = xml.read_element(ename, evalue)	' </DDSDirector>
	ret = xml.read_element(ename, evalue)	' </sr2chumktool>
	xml.close_doc()

	return(F_NOERROR)
end function
