'----------------------------------------------------------
'	SR2 Chunks Texture Replacement Tool
'	Copyright (c) 2020 Masamichi Kirikae
'----------------------------------------------------------
#include once "windows.bi"
#include once "win/winuser.bi"
#include once "win/commctrl.bi"
#include once "win/commdlg.bi"
#include once "win/shellapi.bi"

#include once "dir.bi" 	'provides constants to use for the attrib_mask parameter
#include once "file.bi"
#include once "fbthread.bi"
#include once "crt.bi"
#include once "datatype.bi"

#include once "fileio.bi"
#include once "pegfile.bi"
#include once "mipgen.bi"
#include once "squish.bi"
#include once "ddslib.bi"
#include once "sr2packfile.bi"
#include once "pnglib.bi"
#include once "xml.bi"
#include once "winfilefolder.bi"
#include once "libstring.bi"
#include once "keycode.bi"

#ifndef WINDOWS_GUI
	#define WINDOWS_GUI
#endif

#ifndef F_NOERROR
	#define	F_NOERROR	0
	#define	F_ERROR		1
#endif

#ifndef NULL
	#define NULL	0
#endif

#ifndef rgba_data
	type rgba_data
		r	as u8
		g	as u8
		b	as u8
		a	as u8
	end type
#endif

#macro exit_main(a)
	cprint ""
	cprint "press any key or close window."
	chdir(currpath)
	close
	sleep
	return (a)
#endmacro

'	Windows GUI ID
#define IDD_DLG1			1000
#define IDC_LST1			1001
#define IDC_LST2			1002
#define IDC_PGB1			1003
#define IDD_DLGABOUT		3000

#define IDR_MENU1			10000
#define IDM_FILE_OPEN		10001
#define IDM_FILE_COPY		10009
#define IDM_FILE_INIT		10007
#define IDM_FILE_EXIT		10002
#define IDM_EDIT_TEMP		10003
#define IDM_MAKE_EXTRACT	10004
#define IDM_MAKE_REBUILD	10005
#define IDM_MAKE_STAT		10008
#define IDM_MAKE_ABORT		10010
#define IDM_HELP_ABOUT		10006

#define ERROR_WRONG_INIFILE		1
#define ERROR_NO_SR2FOLDER		2
#define ERROR_NO_TEMPFOLDER		4
#define ERROR_NO_VPPSOURCE		8
#define ERROR_NO_CHUNKSDATA		16
#define ERROR_MAXVAL			32

#define THREAD_NOT_IN_USE			1
#define THREAD_IN_USE				2


declare function AboutDlgProc(byval hWin as HWND,byval uMsg as UINT,byval wParam as WPARAM,byval lParam as LPARAM) as s32
declare function DialogProc(byval hWin as HWND, byval uMsg as UINT, ByVal wParam as WPARAM, byval lParam as LPARAM) As s32

#define	PROGRAM_VERSION	"1.271"
#define	_INIFILE_NAME	"SR2ChunksTextureTool.ini"
#define	_DEBUG		0

'	chunktool
#define _ORIGINAL_VPP_PC	"original_vpp_pc"
#define _REBUILD_VPP_PC		"new_vpp_pc"
#define _TEMP_CHUNKS		"\chunks"
#define _TEMP_DDS			"\dds"
#define _TEMP_PEGS			"\pegs"
#define _TOTAL_CHUNKS		385

#define MAX_TEXTURES 		15000
#define MAX_NEW_TEXTURES 	512
#define MAX_CHUNKS			150

type chunk_data
	name as string
	replaceflag as s32
end type

type chunks_data
	num as s32
	chunk(MAX_CHUNKS) as chunk_data
end type

type texture_data
	name as string
	replaceflag as s32
end type

type replaced_texture_data
	index as s32
end type

type chunk_textures_data
	name as string
	chunkname as string
	width as s16
	height as s16
	format as s16
	num as s16
end type

type MessageStack
	n as s32
	message as string
end type

dim shared as string rootpath
dim shared as string temppath
dim shared as string currpath
dim shared as string sr2path
dim shared as s32 total_chunks  = 0
dim shared as chunks_data chunks(2)	' 0,1,2
dim shared as texture_data textures(MAX_NEW_TEXTURES)

dim shared processed_textures as s32 = 0
dim shared replacedn as s32 = 0
dim shared replaced(MAX_NEW_TEXTURES) as replaced_texture_data

dim shared hInstance as HMODULE
dim shared _hDlg as HWND
dim shared as HMENU hMenu
dim shared hIcon as HICON
dim shared hProg as HWND
dim shared hLBox as HWND
dim shared Prog_Max as u32
dim shared Prog_Val as u32

dim shared as s32 THREAD_STATUS = THREAD_NOT_IN_USE
dim shared as s32 THREAD_RETURN = FALSE
dim shared as s32 THREAD_ERROR = FALSE
dim shared as s32 THREAD_ABORT = FALSE

dim shared as MessageStack ListStack(80)
dim	shared as s32 StackDepth = 0

declare sub abort_thread()
declare sub delete_temporary_files()
declare sub delete_temporary_subfolders()
declare sub AddMessage(byref st as zstring)
declare sub DelMessage(byref st as zstring)
declare sub display_time()
declare function check_vpp_pc_files(path as string) as s32
declare sub copy_vpp_pc_sourcefiles()
declare sub copy_vpp_pc_newfiles()
declare function compare_image(srcfile as string, dstfile as string) as double
declare function delete_file(destpath as string, attrib_mask as s32) as s32
declare function copy_file(sourcepath as string, destpath as string, attrib_mask as s32) as s32
declare function move_onefile(sourcepath as string, destpath as string) as s32
declare function drive_exist(path as string) as s32
declare sub init_chunks_data()
declare function init_temporary_folder() as s32
declare function get_image_colour_string(n as s32) as string
declare function get_dds_settings() as string
declare function write_inifile(filename as string) as s32
declare function read_inifile(filename as string) as s32
declare function get_trash_textures() as s32
declare function get_new_textures() as s32
declare function get_chunk_names_from_file(sourcefile as string) as s32
declare function get_chunk_names() as s32
declare function restore_original_textures(texn as s32) as s32
declare function replace_textures(cn as s32, texn as s32) as s32
declare sub replace_textures_and_rebuild()
declare sub make_chunks_texture_data()
declare sub make_texture_statistics()