'----------------------------------------------------------
'	DDS File load and save functions
'	Copyright (c) 2020 Masamichi Kirikae
'----------------------------------------------------------
#include once "ddslib.bi"

namespace dds
	
'--------------------------------------------------------------------
'	initialize dds library director variables
'--------------------------------------------------------------------
sub init_dds()
	dds_director.silentmode = false
	dds_director.render_target = DDS_SR2_ENGINE
	dds_director.elapsed_time = 0
	'  squish
	dds_director.colourfit = squish_kColourClusterFit
	'	DXT1
	dds_director.mipmap(DDS_MIPMAP_DXT1).resize_filter = "BOX_DITHER_DIFFUSION"
	dds_director.mipmap(DDS_MIPMAP_DXT1).colourspace = DDS_sRGB_COLOUR
	'	DXT3, DXT5
	dds_director.mipmap(DDS_MIPMAP_DXT3_DXT5).resize_filter = "BOX_DITHER_DIFFUSION"
	dds_director.mipmap(DDS_MIPMAP_DXT3_DXT5).colourspace = DDS_sRGB_COLOUR
	'	NORMALMAP
	dds_director.mipmap(DDS_MIPMAP_NORMALMAP).resize_filter = "BOX"
	dds_director.mipmap(DDS_MIPMAP_NORMALMAP).colourspace = DDS_LINEAR_COLOUR
end sub

'--------------------------------------------------------------------
'	initialize dds header by width, height, pixelformat, mipmaplevel
'	return value mipmapcount:success 0:Error
'--------------------------------------------------------------------
function init_ddsheader(byref ddshdr as struct_ddsheader, iwidth as s32, iheight as s32, fmt as s32, mipmapcount as s32) as s32
	dim i as s32, flags as s32

	ddshdr.dwMagic = &H20534444
	ddshdr.dwSize  = 124

	select case fmt
		case D3DFMT_A8R8G8B8
			'mipmap level = 1
			ddshdr.dwFlags = DDSD_CAPS or DDSD_HEIGHT or DDSD_WIDTH or DDSD_PIXELFORMAT
			ddshdr.dwPitchOrLinear = iwidth * 4	' bytes/row
		case D3DFMT_R5G6B5
			'mipmap level = 1
			ddshdr.dwFlags = DDSD_CAPS or DDSD_HEIGHT or DDSD_WIDTH or DDSD_PIXELFORMAT
			ddshdr.dwPitchOrLinear = iwidth * 2	' bytes/row
		case D3DFMT_DXT1 
			if (mipmapcount > 1) then
				ddshdr.dwFlags = DDSD_CAPS or DDSD_HEIGHT or DDSD_WIDTH or DDSD_PIXELFORMAT or DDSD_MIPMAPCOUNT or DDSD_LINEARSIZE
			else
				ddshdr.dwFlags = DDSD_CAPS or DDSD_HEIGHT or DDSD_WIDTH or DDSD_PIXELFORMAT or DDSD_LINEARSIZE
			end if
			flags = squish_kDxt1
			ddshdr.dwPitchOrLinear = squish.GetStorageRequirements(iwidth, iheight, flags)
		case D3DFMT_DXT3
			if (mipmapcount > 1) then
				ddshdr.dwFlags = DDSD_CAPS or DDSD_HEIGHT or DDSD_WIDTH or DDSD_PIXELFORMAT or DDSD_MIPMAPCOUNT or DDSD_LINEARSIZE
			else
				ddshdr.dwFlags = DDSD_CAPS or DDSD_HEIGHT or DDSD_WIDTH or DDSD_PIXELFORMAT or DDSD_LINEARSIZE
			end if
			flags = squish_kDxt3
			ddshdr.dwPitchOrLinear = squish.GetStorageRequirements(iwidth, iheight, flags)
		case D3DFMT_DXT5
			if (mipmapcount > 1) then
				ddshdr.dwFlags = DDSD_CAPS or DDSD_HEIGHT or DDSD_WIDTH or DDSD_PIXELFORMAT or DDSD_MIPMAPCOUNT or DDSD_LINEARSIZE
			else
				ddshdr.dwFlags = DDSD_CAPS or DDSD_HEIGHT or DDSD_WIDTH or DDSD_PIXELFORMAT or DDSD_LINEARSIZE
			end if
			flags = squish_kDxt5
			ddshdr.dwPitchOrLinear = squish.GetStorageRequirements(iwidth, iheight, flags)
		case else
			cprint "Error: unsupported format"
			return(0)	' Error
	end select

	ddshdr.dwHeight			= iheight
	ddshdr.dwWidth			= iwidth
	ddshdr.dwDepth			= 0	' dwDepth for volume Texture
	ddshdr.dwMipMapCount	= mipmapcount

	for i = 0 to 10
		ddshdr.dwReserved1(i) = 0		' reserved1
	next
	ddshdr.dwReserved1(9) = &H54474550	' TGEP	(PEGTool)
	
	ddshdr.dwPfSize = 32
	
	select case fmt
		case D3DFMT_A8R8G8B8
			ddshdr.dwPfFlags = DDPF_RGB or DDPF_ALPHAPIXELS
			ddshdr.dwFourCC = 0
			ddshdr.dwRGBBitCount = 32
			ddshdr.dwRBitMask = &H00FF0000
			ddshdr.dwGBitMask = &H0000FF00
			ddshdr.dwBBitMask = &H000000FF
			ddshdr.dwABitMask = &HFF000000
			ddshdr.dwCaps1 = DDSCAPS_TEXTURE	' DDSCAPS_MIPMAP, DDSCAPS_COMPLEX (add these flags when make mipmap with A8R8G8B8)
		case D3DFMT_R5G6B5
			ddshdr.dwPfFlags = DDPF_RGB
			ddshdr.dwFourCC = 0
			ddshdr.dwRGBBitCount = 16
			ddshdr.dwRBitMask = &HF800
			ddshdr.dwGBitMask = &H07E0
			ddshdr.dwBBitMask = &H001F
			ddshdr.dwCaps1 = DDSCAPS_TEXTURE	' DDSCAPS_MIPMAP, DDSCAPS_COMPLEX (add these flags when make mipmap with R5G6B5)
		case D3DFMT_DXT1
			ddshdr.dwPfFlags = DDPF_FOURCC
			ddshdr.dwFourCC = &H31545844		' 1TXD
			ddshdr.dwRGBBitCount = 0
			ddshdr.dwRBitMask = 0
			ddshdr.dwGBitMask = 0
			ddshdr.dwBBitMask = 0
			ddshdr.dwABitMask = 0
			if (mipmapcount > 1) then
				ddshdr.dwCaps1 = DDSCAPS_TEXTURE or DDSCAPS_MIPMAP or DDSCAPS_COMPLEX
			else
				ddshdr.dwCaps1 = DDSCAPS_TEXTURE
			end if
		case D3DFMT_DXT3
			ddshdr.dwPfFlags = DDPF_FOURCC
			ddshdr.dwFourCC = &H33545844		' 3TXD
			ddshdr.dwRGBBitCount = 0
			ddshdr.dwRBitMask = 0
			ddshdr.dwGBitMask = 0
			ddshdr.dwBBitMask = 0
			ddshdr.dwABitMask = 0
			if (mipmapcount > 1) then
				ddshdr.dwCaps1 = DDSCAPS_TEXTURE or DDSCAPS_MIPMAP or DDSCAPS_COMPLEX
			else
				ddshdr.dwCaps1 = DDSCAPS_TEXTURE
			end if
		case D3DFMT_DXT5
			ddshdr.dwPfFlags = DDPF_FOURCC
			ddshdr.dwFourCC = &H35545844		' 5TXD
			ddshdr.dwRGBBitCount = 0
			ddshdr.dwRBitMask = 0
			ddshdr.dwGBitMask = 0
			ddshdr.dwBBitMask = 0
			ddshdr.dwABitMask = 0
			if (mipmapcount > 1) then
				ddshdr.dwCaps1 = DDSCAPS_TEXTURE or DDSCAPS_MIPMAP or DDSCAPS_COMPLEX
			else
				ddshdr.dwCaps1 = DDSCAPS_TEXTURE
			end if
		case else
			cprint "Error:unsupported format"
			return(0)	' Error. unsupported format
	end select

	ddshdr.dwCaps2 = 0		' for cubemap or volume texutre flag etc.
	ddshdr.Reserved(0) = 0
	ddshdr.Reserved(1) = 0	' last element of ddsCaps structure.
	ddshdr.dwReserved2 = 0	' last element of ddsheader structure. 
	
	return(mipmapcount) ' No error
	
end function

'--------------------------------------------------------------------
' load dds header only
' return mipmapcount
'--------------------------------------------------------------------
function load_ddsheader(byref fname as string, byref imgw as s32, byref imgh as s32, byref pixform as s32) as s32
	dim ddshdr as struct_ddsheader
	dim as s32 fh, i, n, x, y, datasize
	dim c16 as u16, crgba as rgba_data

	fh = freefile
	open fname for binary access read as #fh
	if (err) then
		cprint "Error:can not open file."
		close #fh
		return(0)
	endif
	
	get #fh, , ddshdr.dwMagic				' = 0x20534444
	if (ddshdr.dwMagic <> &H20534444) then	' check dwMagic is " SDD"
		cprint "Error:this file is not dds format."
		close #fh
		return(0)
	end if
	get #fh, , ddshdr.dwSize			' = 124
	get #fh, , ddshdr.dwFlags			' valid field flag. contains DDSD_CAPS, DDSD_PIXELFORMAT, DDSD_WIDTH, DDSD_HEIGHT
	get #fh, , ddshdr.dwHeight	
	get #fh, , ddshdr.dwWidth
	get #fh, , ddshdr.dwPitchOrLinear	' PitchOrLinearSize (pitch:bytes/line LinearSize:bytes/Frame
	get #fh, , ddshdr.dwDepth			' for volume Texture
	get #fh, , ddshdr.dwMipMapCount		' mipmap count
	for i = 0 to 10
	  get #fh, , ddshdr.dwReserved1(i)	' reserved 11 DWORD (0 to 10) value is 0	
	next
	get #fh, , ddshdr.dwPfSize			' = 32
	get #fh, , ddshdr.dwPfFlags			' DDPF_RGB / DDPF_FOURCC
	get #fh, , ddshdr.dwFourCC			' set value when dwPfFlags is DDPF_FOURCC
	get #fh, , ddshdr.dwRGBBitCount		' bits/pixel with RGB format (must be 0 when DXT1-5)
	get #fh, , ddshdr.dwRBitMask		' R mask with RGB Format (when DDPF_RGB)
	get #fh, , ddshdr.dwGBitMask		' G mask with RGB Format (when DDPF_RGB)
	get #fh, , ddshdr.dwBBitMask		' B mask with RGB Format (when DDPF_RGB)
	get #fh, , ddshdr.dwABitMask		' alpha mask with RGBA Format (when DDPF_RGB)
	get #fh, , ddshdr.dwCaps1			' DDSD_CAPS_TEXTURE(must), DDSD_CAPS_MIPMAP | DDSD_CAPS_COMPLEX (size of ddsCaps structure is 16byte)
	get #fh, , ddshdr.dwCaps2			' for cubemap or volume texutre flag etc.
	get #fh, , ddshdr.Reserved(0)		' reserved DWORD (0 to 1)  last elemet of ddsCaps structure
	get #fh, , ddshdr.Reserved(1)
	get #fh, , ddshdr.dwReserved2		' reserved DWORD
	
	imgw = cast(s32, ddshdr.dwWidth)
	imgh = cast(s32, ddshdr.dwHeight)
	pixform = 0
	select case ddshdr.dwPfFlags
		case (DDPF_RGB or DDPF_ALPHAPIXELS)		'R8G8B8A8 BGRA in file
			if (ddshdr.dwRGBBitCount = 32) then
				pixform = D3DFMT_A8R8G8B8
			else
				cprint "Error:dwRGBBitCount value is wrong."
				close #fh
				return(0)
			end if
		case DDPF_RGB
			if (ddshdr.dwRGBBitCount = 16) then
				pixform = D3DFMT_R5G6B5
			else
				cprint "Error:dwRGBBitCount value is wrong."
				close #fh
				return(0)
			end if
		case DDPF_FOURCC
			select case ddshdr.dwFourCC
				case &H31545844		' 1TXD
					pixform = D3DFMT_DXT1
				case &H33545844		' 3TXD
					pixform = D3DFMT_DXT3
				case &H35545844		' 5TXD
					pixform = D3DFMT_DXT5
				case else
					cprint "Error:not supported pixel format."
					close #fh
					return(0)
			end select
		case else
			cprint "Error:pixel format infomation in dds header is wrong."
			close #fh
			return(0)
	end select

	close #fh	
	return(cast(s32, ddshdr.dwMipMapCount))

end function

'--------------------------------------------------------------------
' load dds file
' image() contains rgba data when pixform is A8R8G8B8.
' image() contains compressed data with mipmap when pixform is DXT1,3,5.
' error returns 0. no error returns mipmapcount.
'--------------------------------------------------------------------
function load_dds(byref fname as string, image() as u8, byref imgw as s32, byref imgh as s32, byref pixform as s32) as s32
	dim ddshdr as struct_ddsheader
	dim as s32 fh, i, n, x, y, datasize
	dim c16 as u16, crgba as rgba_data
	
	datasize = filelen(fname) - 128
	
	fh = freefile
	open fname for binary access read as #fh
	if (err) then
		cprint "Error:can not open file."
		close #fh
		return(0)
	endif
	
	get #fh, , ddshdr.dwMagic				' = 0x20534444
	if (ddshdr.dwMagic <> &H20534444) then	' check dwMagic is " SDD"
		cprint "Error:this file is not dds format."
		close #fh
		return(0)
	end if
	get #fh, , ddshdr.dwSize			' = 124
	get #fh, , ddshdr.dwFlags			' valid field flag. contains DDSD_CAPS, DDSD_PIXELFORMAT, DDSD_WIDTH, DDSD_HEIGHT
	get #fh, , ddshdr.dwHeight
	get #fh, , ddshdr.dwWidth
	get #fh, , ddshdr.dwPitchOrLinear	' PitchOrLinearSize (pitch:bytes/line LinearSize:bytes/Frame
	get #fh, , ddshdr.dwDepth			' for volume Texture
	get #fh, , ddshdr.dwMipMapCount		' store mipmap level
	for i = 0 to 10
	  get #fh, , ddshdr.dwReserved1(i)	' reserved 11 DWORD (0 to 10) value is 0	
	next
	get #fh, , ddshdr.dwPfSize			' = 32
	get #fh, , ddshdr.dwPfFlags			' DDPF_RGB / DDPF_FOURCC
	get #fh, , ddshdr.dwFourCC			' set value when dwPfFlags is DDPF_FOURCC
	get #fh, , ddshdr.dwRGBBitCount		' bits/pixel with RGB format (must be 0 when DXT1-5)
	get #fh, , ddshdr.dwRBitMask		' R mask with RGB Format (when DDPF_RGB)
	get #fh, , ddshdr.dwGBitMask		' G mask with RGB Format (when DDPF_RGB)
	get #fh, , ddshdr.dwBBitMask		' B mask with RGB Format (when DDPF_RGB)
	get #fh, , ddshdr.dwABitMask		' alpha mask with RGBA Format (when DDPF_RGB)
	get #fh, , ddshdr.dwCaps1			' DDSD_CAPS_TEXTURE(must), DDSD_CAPS_MIPMAP | DDSD_CAPS_COMPLEX (size of ddsCaps structure is 16byte)
	get #fh, , ddshdr.dwCaps2			' for cubemap or volume texutre flag etc.
	get #fh, , ddshdr.Reserved(0)		' reserved DWORD (0 to 1)  last elemet of ddsCaps structure
	get #fh, , ddshdr.Reserved(1)
	get #fh, , ddshdr.dwReserved2		' reserved DWORD
	
	imgw = cast(s32, ddshdr.dwWidth)
	imgh = cast(s32, ddshdr.dwHeight)
	pixform = 0
	select case ddshdr.dwPfFlags
		case (DDPF_RGB or DDPF_ALPHAPIXELS)		'R8G8B8A8 BGRA in file
			if (ddshdr.dwRGBBitCount = 32) then
				pixform = D3DFMT_A8R8G8B8
			else
				cprint "Error:dwRGBBitCount value is wrong."
				close #fh
				return(0)
			end if
		case DDPF_RGB
			if (ddshdr.dwRGBBitCount = 16) then
				pixform = D3DFMT_R5G6B5
			else
				cprint "Error:dwRGBBitCount value is wrong."
				close #fh
				return(0)
			end if
		case DDPF_FOURCC
			select case ddshdr.dwFourCC
				case &H31545844		' 1TXD
					pixform = D3DFMT_DXT1
				case &H33545844		' 3TXD
					pixform = D3DFMT_DXT3
				case &H35545844		' 5TXD
					pixform = D3DFMT_DXT5
				case else
					cprint "Error:not supported pixel format."
					close #fh
					return(0)
			end select
		case else
			cprint "Error:pixel format infomation in dds header is wrong."
			close #fh
			return(0)
	end select

	fpseek(fh, 128)
	select case pixform
		case D3DFMT_A8R8G8B8
			n = 0
			for y = 0 to imgh - 1
				for x = 0 to imgw - 1
					get #fh, ,crgba.b
					get #fh, ,crgba.g
					get #fh, ,crgba.r
					get #fh, ,crgba.a
					image(n) = crgba.r: n+=1
					image(n) = crgba.g: n+=1
					image(n) = crgba.b: n+=1
					image(n) = crgba.a: n+=1
				next
			next
		case D3DFMT_R5G6B5
			datasize = imgh * imgw * 4		' convert R5G6B5 to r8g8b8a8
			n = 0
			for y = 0 to imgh - 1
				for x = 0 to imgw - 1
					get #fh, ,c16	' u16 R5G6B5
					rgb565_to_rgba(c16, crgba.r, crgba.g, crgba.b, crgba.a)
					image(n) = crgba.r: n+=1
					image(n) = crgba.g: n+=1
					image(n) = crgba.b: n+=1
					image(n) = crgba.a: n+=1
				next
			next
		case D3DFMT_DXT1, D3DFMT_DXT3, D3DFMT_DXT5
			get #fh, ,image(0) ,datasize
		case else
	end select

	close #fh
	return(cast(s32, ddshdr.dwMipMapCount))

end function

sub printError(byval message as const zstring ptr)
	dim as LPVOID lpvMessageBuffer
	
	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER or FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), lpvMessageBuffer, 0, NULL)
	cprint *message, *cast(zstring ptr, lpvMessageBuffer)
	LocalFree(LpvMessageBuffer)
end sub		

'--------------------------------------------------------------------
'	save_dds_rgba() makes dds file from uncompressed rgba format image.
'	return value mipmapcount(>0):success 0:error 
'--------------------------------------------------------------------
function save_dds_rgba(byref fname as string, image_rgba() as u8, imgw as s32, imgh as s32, pixform as s32, mipmapflag as s32) as s32
	dim ddshdr as struct_ddsheader
	dim as s32 imagew, imageh, size, Size_DxtCompressed
	dim as s32 fh, flags, i, x, y, n
	dim as s32 mip, max_mip, mipmapcount
	dim as rgba_data crgba
	dim as u16 c16
	dim as s32 flag_normalmap
	dim as double start_time, end_time

	if ((imgw < 1) or (imgh < 1)) then return(0)
	
	start_time = timer	' for benchmark
	
	select case pixform
		case D3DFMT_A8R8G8B8
			fh = freefile
			open fname for output as #fh
			mipmapcount = 1
			init_ddsheader(ddshdr, imgw, imgh, pixform, mipmapcount)
			put #fh, ,ddshdr
			i = 0
			for y = 0 to imgh - 1
				for x = 0 to imgw - 1
					crgba.r = image_rgba(i): i+=1
					crgba.g = image_rgba(i): i+=1
					crgba.b = image_rgba(i): i+=1
					crgba.a = image_rgba(i): i+=1
					put #fh, ,crgba.b
					put #fh, ,crgba.g
					put #fh, ,crgba.r
					put #fh, ,crgba.a
				next
			next
			close #fh
			return(mipmapcount)	' Success.
		case D3DFMT_R5G6B5
			fh = freefile
			open fname for output as #fh
			mipmapcount = 1
			init_ddsheader(ddshdr, imgw, imgh, pixform, mipmapcount)
			put #fh, ,ddshdr
			i = 0
			for y = 0 to imgh - 1
				for x = 0 to imgw - 1
					crgba.r = image_rgba(i): i+=1
					crgba.g = image_rgba(i): i+=1
					crgba.b = image_rgba(i): i+=1
					crgba.a = image_rgba(i): i+=1
					rgb_to_rgb565(c16, crgba.r, crgba.g, crgba.b)
					put #fh, ,c16
				next
			next
			close #fh
			return(mipmapcount)	' Success.
		case D3DFMT_DXT1, D3DFMT_DXT3, D3DFMT_DXT5
			'	check image with and height is 2^n (n > 2)
			if (check_dxtc_sourceimagesize(imgw, imgh, mipmapflag) = false) then
				cprint "image_size_error "; mipmapflag
				return(0)
			endif
			' set mipmaplevel
			select case mipmapflag
				case 101
					' no mipmap
					mipmapcount = 1
				case 1 to 99
					' auto mipmap level
					max_mip = get_maxmipmaplevel(imgw, imgh)
					mipmapcount = max_mip
				case is > 101
					' mipmap level restriction
					max_mip = get_maxmipmaplevel(imgw, imgh)
					mipmapcount = mipmapflag - 100
					if (mipmapcount > max_mip) then mipmapcount = max_mip
				case else
					mipmapcount = 1
			end select
		case else
			return(0)	' Error. unsupported format
	end select

	'	set DXT format and compress method flag
	select case pixform
		case D3DFMT_DXT1
			flags = squish_kDxt1 or dds_director.colourfit
		case D3DFMT_DXT3
			flags = squish_kDxt3 or dds_director.colourfit
		case D3DFMT_DXT5
			flags = squish_kDxt5 or dds_director.colourfit
		case else
	end select

	dim metric as float ptr
	dim as string resizefilter, colourstring

	flag_normalmap = is_normalmap(image_rgba(), imgw, imgh, pixform)
	
	'	set resize method, colourmetric and colourspace
	if (flag_normalmap) then	'	normalmap exception
		resizefilter = " -if " + dds_director.mipmap(DDS_MIPMAP_NORMALMAP).resize_filter
		select case dds_director.mipmap(DDS_MIPMAP_NORMALMAP).colourspace
			case DDS_sRGB_COLOUR
				colourstring = " -srgb"
			case DDS_LINEAR_COLOUR
				colourstring = ""
			case else
		end select
	else
		select case pixform
			case D3DFMT_DXT1
				resizefilter = " -if " + dds_director.mipmap(DDS_MIPMAP_DXT1).resize_filter
				select case dds_director.mipmap(DDS_MIPMAP_DXT1).colourspace
					case DDS_sRGB_COLOUR
						colourstring = " -srgb"
					case DDS_LINEAR_COLOUR
						colourstring = ""
					case else
				end select
			case D3DFMT_DXT3, D3DFMT_DXT5
				flags = flags or squish_kWeightColourByAlpha	'	squish compressor consider alpha
				resizefilter = " -if " + dds_director.mipmap(DDS_MIPMAP_DXT3_DXT5).resize_filter
				select case dds_director.mipmap(DDS_MIPMAP_DXT3_DXT5).colourspace
					case DDS_sRGB_COLOUR
						colourstring = " -srgb"
					case DDS_LINEAR_COLOUR
						colourstring = ""
					case else
				end select
			case else
		end select
	endif
	
	if (not(dds_director.silentmode)) then
		cprint using "####x#### mipmap ## compressing..."; imgw; imgh; mipmapcount;
	endif
	
	if (mipmapcount = 1) then
		
		size = imgw * imgh * 4
		arraycopy(squish.image_rgba, image_rgba, size)

		if (colourstring <> "" ) then
			metric = @squish.metric_perceptual(0)
		else
			metric = @squish.metric_uniform(0)
		endif
		
		'	write dds header
		fh = freefile
		open fname for output as #fh
		
		init_ddsheader(ddshdr, imgw, imgh, pixform, mipmapcount)
		put #fh, ,ddshdr
		
		Size_DxtCompressed = squish.GetStorageRequirements(imgw, imgh, flags)
		squish.CompressImage(@squish.image_rgba(0), imgw, imgh, @squish.dxtblocks(0), flags, metric)
		
		put #fh, ,squish.dxtblocks(0), Size_DxtCompressed
		close #fh

	else
		
		dim as zstring*512 cmdstring
		dim as string fmtstring, colourstring, outdir, srcfile, mipstr

		mipstr = " -m "+str(mipmapcount)
		select case pixform
			case D3DFMT_DXT1
				fmtstring = "DXT1"
			case D3DFMT_DXT3
				fmtstring = "DXT3"
			case D3DFMT_DXT5
				fmtstring = "DXT5"
			case else
		end select

		srcfile = libfio.change_filename_ext(fname, ".png") 
		outdir = rtrim(libfio.get_directory(fname),"\")	

		cmdstring = "texconv.exe " + mipstr + " -nologo -nowic -y -f " + fmtstring + resizefilter + colourstring + _
		" -ft dds -wrap -nogpu -fl 9.3 -o "	+ !"\"" + outdir + !"\" " + !"\"" + srcfile + !"\""
	
		'shell(cmdstring)

		dim as STARTUPINFO si
		dim as PROCESS_INFORMATION pi
		
		si.cb = sizeof(STARTUPINFO)
				
		flags = CreateProcess(NULL, cast(zstring ptr, @cmdstring), NULL, NULL, FALSE, DETACHED_PROCESS, NULL, NULL, @Si, @Pi)
		if (flags = 0) then
			cprint("CreateProcess Fail.")
			sleep
			return(0)
		else
			WaitForSingleObject(pi.hProcess, INFINITE)
		    CloseHandle(pi.hProcess)
		    CloseHandle(pi.hThread)
		endif
		
		srcfile = libfio.change_filename_ext(fname, ".dds")
		if (dir(fname, fbArchive) <> "") then
			Name(fname, fname)
		endif
			
		
	endif

	end_time = timer
	dds_director.elapsed_time = end_time - start_time

	if (not(dds_director.silentmode)) then
		cprint "Done."
	endif

	return(mipmapcount)	' success
end function

'--------------------------------------------------------------------
'	dxtc image width and height must be 2^n (n > 2) with mipmap
'  or 4*n (n>0) without mipmap
'--------------------------------------------------------------------
function check_dxtc_sourceimagesize(w as s32, h as s32, mip as s32) as s32
	dim as s32 n
	if (mip <> 101) then
		n = 2
		do
			n = (n shl 1)
			if (w < n) then
				return(false)
			elseif (w = n) then
				exit do
			endif
		loop
		n = 2
		do
			n = (n shl 1)
			if (h < n) then
				return(false)
			elseif (h = n) then
				exit do
			endif
		loop
	else
		if (w mod 4) then return(false)
		if (h mod 4) then return(false)
	endif
	return(true)
end function

'--------------------------------------------------------------------
'	calculate maximum mipmaplevel
'--------------------------------------------------------------------
function get_maxmipmaplevel(imagew as s32, imageh as s32) as s32
	dim as integer max_mip, n, minwidth
	minwidth = dds_director.render_target
	max_mip = 0
	if (imagew < imageh) then
		n = imagew
	else
		n = imageh
	endif
	do until(n < minwidth)
		max_mip += 1
		n = n \ 2
	loop
	return(max_mip)
end function

/'
'--------------------------------------------------------------------
'	RGBA to YCbCrA (IYU-R BT,709)
'--------------------------------------------------------------------
sub convert_RGB_to_YCbCr(byref r as double, byref g as double, byref b as double, byref Y as double, byref Cb as double, byref Cr as double)
	Y  =  0.2126   * r + 0.7152   * g + 0.0772   * b
	Cb = -0.114572 * r - 0.385428 * g + 0.5      * b
	Cr =  0.5      * r - 0.454153 * g - 0.045847 * b
end sub

'--------------------------------------------------------------------
'	YCbCrA (IYU-R BT,709) to RGBA
'--------------------------------------------------------------------
sub convert_YCbCr_to_RGB(byref r as double, byref g as double, byref b as double, byref Y as double, byref Cb as double, byref Cr as double)
	r = Y + 1.5748   * Cr
	g = Y - 0.187324 * Cb - 0.468124 * Cr
	b = Y + 1.8556   * Cb	
end sub

'--------------------------------------------------------------------
'	sRGB to Linear colourspace
'--------------------------------------------------------------------
function sRGB_to_linear(cs as double) as double
	dim as double cl
	if (cs <= 0.04045) then
		cl = cs / 12.92
	else
		cl =  ((cs + 0.055) / 1.055)^2.4
	endif
	return(cl)
end function

'--------------------------------------------------------------------
'	Linear to sRGB colourspace
'--------------------------------------------------------------------
function Linear_to_sRGB(c as double) as double
	dim as double dc
	if (c <= 0.0031308) then
		dc = 12.92 * c
	else
		dc = 1.055 * c^(1/2.4) - 0.055
	endif
	return(dc)
end function

'--------------------------------------------------------------------
'	first round for fast sin
'--------------------------------------------------------------------
function fast_round(byval x as double) as long
   dim MAGIC_ROUND as const double = 6755399441055744.0
   union fast_trunc
      d as double
      type
         lw as long
         hw as long
      end type
   end union
   dim fast_trunc as fast_trunc
   fast_trunc.d = x
   fast_trunc.d += MAGIC_ROUND
   return fast_trunc.lw
end function

'--------------------------------------------------------------------
'	fast sin function
'--------------------------------------------------------------------
function fast_sin(byval x as double) as double
   dim FPI as const double   = 3.14159265358979323846264338327950288
   dim INVPI as const double = 0.31830988618379067153776752674502872
   dim A as const double     = 0.00735246819687011731341356165096815
   dim B as const double    = -0.16528911397014738207016302002888890
   dim C as const double     = 0.99969198629596757779830113868360584
   dim k as long
   dim x2 as double
   k = fast_round(INVPI * x)
   x -= k * FPI
   x2 = x * x
   x = x * (C + (x2 * (B + (A * x2))))
   if k mod 2 then
      x = -x
   end if
   return x
end function
'/

'--------------------------------------------------------------------
'	calculate psnr 
'--------------------------------------------------------------------
function get_psnr(ref_rgba() as u8, cmp_rgba() as u8, imgw as s32, imgh as s32) as double
	dim as s32 n, imgsize
	dim as double pixmax, mse, psnr
	dim as rgba_data ref, cmp
	
	imgsize = imgh * imgw * 4
	n = 0
	mse = 0.0
	do
		ref.r = ref_rgba(n)
		ref.g = ref_rgba(n + 1)
		ref.b = ref_rgba(n + 2)
		ref.a = ref_rgba(n + 3)
		
		cmp.r = cmp_rgba(n)
		cmp.g = cmp_rgba(n + 1)
		cmp.b = cmp_rgba(n + 2)
		cmp.a = cmp_rgba(n + 3)

		mse += ( (ref.r - cmp.r)^2.0 + (ref.g - cmp.g)^2.0 + (ref.b - cmp.b)^2.0 + (ref.a - cmp.a)^2.0 )
		n += 4
	loop while (n < imgsize)

	'	calcurate PSNR
	pixmax = imgw * imgh * 4.0
	mse = mse / pixmax
	if (mse > 0.0) then
		psnr = 10.0 * log(65025.0 / mse) / log(10.0)
	else
		psnr = 120.0	' actually no loss
	endif
	return(psnr)
end function

'--------------------------------------------------------------------
' calculate psnr value of dxt3 and dxt5 compression and return most suitable format 
'--------------------------------------------------------------------
function get_suitable_format(image_rgba() as u8, imgw as s32, imgh as s32) as s32
	dim as s32 a, n, fmt, imgsize, cpsize, flags, aflag
	dim as double psnr3, psnr5, sum

	imgsize = imgw * imgh * 4
	
	' check all pixel's alpha value
	aflag = true
	n = 0
	sum = 0
	do
		sum += image_rgba(n + 3)
		n += 4
	loop while (n < imgsize)
	sum = sum / (imgw * imgh)

	'	select DXT1 when average alpha > 254. (99.9% opaque) 
	'	Some of the opaque image has a less than 255 alpha value. Such a case occurs due to some error.
	if (sum > 254) then
		fmt = D3DFMT_DXT1
		return(fmt)
	endif

	dim metric as float ptr
	metric = @squish.metric_perceptual(0)
	'	dxt3
	arraycopy(squish.image_rgba, image_rgba, imgsize)
	flags = squish_kDxt3
	cpsize = squish.GetStorageRequirements(imgw, imgh, flags)
	cprint "compressing DXT3..."
	squish.CompressImage(@squish.image_rgba(0), imgw, imgh, @squish.dxtblocks(0), flags, metric)
	squish.DecompressImage(@squish.image_rgba(0), imgw, imgh, @squish.dxtblocks(0), flags)

	redim dxt3_rgba(imgsize - 1) as u8
	arraycopy(dxt3_rgba, squish.image_rgba, imgsize)

	'	dxt5
	arraycopy(squish.image_rgba, image_rgba, imgsize)
	flags = squish_kDxt5
	cpsize = squish.GetStorageRequirements(imgw, imgh, flags)
	cprint "compressing DXT5..."
	squish.CompressImage(@squish.image_rgba(0), imgw, imgh, @squish.dxtblocks(0), flags, metric)
	squish.DecompressImage(@squish.image_rgba(0), imgw, imgh, @squish.dxtblocks(0), flags)

	redim dxt5_rgba(imgsize - 1) as u8	
	arraycopy(dxt5_rgba, squish.image_rgba, imgsize)

	' calcurate MSE DXT3 / DXT5
	psnr3 = get_psnr(image_rgba(), dxt3_rgba(), imgw, imgh)
	psnr5 = get_psnr(image_rgba(), dxt5_rgba(), imgw, imgh)

	cprint using "DXT3 PSNR ###.#dB"; psnr3
	cprint using "DXT5 PSNR ###.#dB"; psnr5

	' evaluate PSNR
	' higier PSNR value means better image quality
	if (psnr3 > psnr5) then
		fmt = D3DFMT_DXT3
	else
		fmt = D3DFMT_DXT5
	endif
	
	return(fmt)

end function

'--------------------------------------------------------------------
'	normalmap (opaque) to tangentspace normalmap (transparent)
'--------------------------------------------------------------------
sub normalmap_to_tgn(normal_rgba() as u8, tgn_rgba() as u8, imgw as s32, imgh as s32)
	dim as s32 n, x, y, a
	dim as u8 cr, cg
	n= 0
	for y = 0 to imgh - 1
		for x = 0 to imgw - 1
			' normalmap (opaque image)
			cr = normal_rgba(n)			' R is X axis value
			cg = normal_rgba(n + 1)		' G is Y axis value
													' B is Z axis value
													' alpha is 255													
			' tangentspace normalmap
			tgn_rgba(n) = cr: n+=1		' R is X axis value
			tgn_rgba(n) = cr: n+=1		' G is X axis value
			tgn_rgba(n) = cr: n+=1		' B is X axis value
			tgn_rgba(n) = cg: n+=1		' alpha is Y axis value
		next
	next
end sub

'--------------------------------------------------------------------
'	tgn normalmap (transparent) to normalmap (opaque)
'	return true when source image is tgn and success conversion
'	return false when source image is not tgn.
'--------------------------------------------------------------------
function tgn_to_normalmap(tgn_rgba() as u8, normal_rgba() as u8, imgw as s32, imgh as s32) as s32
	dim as s32 n, x, y, c
	dim as u8 r, g, b, a, max_a, min_a

	n= 0
	for y = 0 to imgh - 1
		for x = 0 to imgw - 1
			' tgn normalmap r,g,b,a (transparent)
			r = tgn_rgba(n)				' R is X Axis
			g = tgn_rgba(n + 1)			' G is X Axis
			b = tgn_rgba(n + 2)			' B is X Axis
			a = tgn_rgba(n + 3)			' alpha is Y axis
			
			' Originally the R, G, B values are the same, but there are cases where they are slightly different,
			' so a threshold is set.
			c = (r + g + b) / 3.0 - r	' R,G,B 
			if (c > 16) then return(false)
			
			' conver to normalmap (opaque)
			normal_rgba(n) = r:		n+=1	' R is X Axis
			normal_rgba(n) = a:		n+=1	' G is Y Axis
			normal_rgba(n) = 255:	n+=1	' B is Z Axis (255 in tangentspace)
			normal_rgba(n) = 255:	n+=1	' alpha is 255
		next
	next	
	return(true)
end function

'--------------------------------------------------------------------
' get colour distance (RGBA)
'--------------------------------------------------------------------
function get_colour_distance(c1 as rgba_data, c2 as rgba_data) as double option ("sse")

	dim as double r1, g1, b1, a1, r2, g2, b2, a2
	dim as double Y1, Cb1, Cr1, Y2, Cb2, Cr2
	'dim as double Yd, Cbd, Crd, Cdiff 
	dim as double rd, gd, bd, ad, Cdiff 
	
	r1 = c1.r / 255.0
	g1 = c1.g / 255.0
	b1 = c1.b / 255.0
	a1 = c1.b / 255.0

	r2 = c2.r / 255.0
	g2 = c2.g / 255.0
	b2 = c2.b / 255.0
	a2 = c2.a / 255.0
	
	rd = (r2 - r1) * (r2 - r1)
	gd = (g2 - g1) * (g2 - g1)
	bd = (b2 - b1) * (b2 - b1)
	ad = (a2 - a1) * (a2 - a1)
	
	Cdiff = sqr(rd + gd + bd + ad) * 255.0
	return(cdiff)
end function

/'
'--------------------------------------------------------------------
' get avarage colour of the image quickly (sample number is imgw * 2)
'--------------------------------------------------------------------
function get_average_colour_fast(image_rgba() as u8, imgw as s32, imgh as s32) as rgba_data
	dim as s32 x, y, n
	dim as s32 r, g, b, a, pixn
	dim as rgba_data c(imgw -1), avrc
	dim as double t, k, s
	
	k = pi * 2.0 / imgw
	pixn = 0
	for x = 0 to imgw - 1
		t = k * x
		s = (fast_sin(t) + 1.0) / 2.0	' normalize 0.0 to 1.0
		y = s * (imgh - 1)
		n = y * imgw * 4 + x * 4
		if (image_rgba(n + 3) > 0) then
			r += image_rgba(n): n+=1
			g += image_rgba(n): n+=1
			b += image_rgba(n): n+=1
			a += image_rgba(n)
			pixn+=1
		else
			'	skip transparent pixel.
			n+=4
		endif
		s = (1.0 - fast_sin(t)) / 2.0	' normalize 0.0 to 1.0
		y = s * (imgh - 1)
		n = y * imgw * 4 + x * 4
		if (image_rgba(n + 3) > 0) then
			r += image_rgba(n): n+=1
			g += image_rgba(n): n+=1
			b += image_rgba(n): n+=1
			a += image_rgba(n)
			pixn+=1
		else
			'	skip transparent pixel.
			n+=4
		endif
	next

	r /= pixn
	g /= pixn
	b /= pixn
	a /= pixn
	
	avrc.r = cubyte(r)
	avrc.g = cubyte(g)
	avrc.b = cubyte(b)
	avrc.a = cubyte(a)
	return(avrc)
end function
'/

'--------------------------------------------------------------------
'	get avarage colour of the image (total inspection)
'--------------------------------------------------------------------
function get_average_colour(image_rgba() as u8, imgw as s32, imgh as s32) as rgba_data
	dim as s64 n, size
	dim as double r, g, b, a
	dim as rgba_data avrc

	size = imgw * imgh * 4
	n = 0		
	do while (n < size)
		r += image_rgba(n): n+=1
		g += image_rgba(n): n+=1
		b += image_rgba(n): n+=1
		a += image_rgba(n): n+=1
	loop
	size \= 4
	r /= size
	g /= size
	b /= size
	a /= size
	avrc.r = cubyte(r)
	avrc.g = cubyte(g)
	avrc.b = cubyte(b)
	avrc.a = cubyte(a)
	return(avrc)
end function

'--------------------------------------------------------------------
'	convert rgb565 to argb8888 with normalize
'--------------------------------------------------------------------
sub rgb565_to_rgba(byref c as u16, byref r as u8, byref g as u8, byref b as u8, byref a as u8)
	dim as u16 id, ir, ig, ib
	ir = (c and &HF800) shr 8		' &b1111 1000 0000 0000
	id = (c and &HE000) shr 12		' &b1110 0000 0000 0000	upper 3 bit to lower 3bit
	ir = ir or id

	ig = (c and &H07E0) shr 3		' &b0000 0111 1110 0000
	id = (c and &H0500) shr 9    	' &b0000 0110 0000 0000
	ig = ig or id

	ib = (c and &H001F) shl 3		' &b0000 0000 0001 1111
	id = (c and &H001C) shr 2		' &b0000 0000 0001 1100
	ib = ib or id
	
	r = cubyte(ir)
	g = cubyte(ig)
	b = cubyte(ib)
	a = &Hff
end sub

'--------------------------------------------------------------------
'	check whether or not the image is a normalmap/tgn
'--------------------------------------------------------------------
function is_normalmap(image_rgba() as u8, imgw as s32, imgh as s32, pixform as s32) as s32
	dim as rgba_data avrcolour, nmpcolour
	dim as s32 size, ret
	dim as double colour_diff
	dim imagebuf_rgba(any) as u8
	' check the texture image whether it is a normal map.
	select case pixform
		case D3DFMT_DXT1
			' there is a possibility of general normalap (opaque)
			avrcolour = get_average_colour(image_rgba(), imgw, imgh)
		case D3DFMT_DXT3, D3DFMT_DXT5
			' There is a possibility of normalmap (transparent) or tgn (transparent)
			size = imgh * imgw * 4
			redim imagebuf_rgba(size - 1)
			ret = tgn_to_normalmap(image_rgba(), imagebuf_rgba(), imgw, imgh)
			if (ret) then
				avrcolour = get_average_colour(imagebuf_rgba(), imgw, imgh)
			else
				' transparental image is not tgn
				avrcolour = get_average_colour(image_rgba(), imgw, imgh)				
			endif
		case else
			return(false)
	end select
	'	flat normalmap's average colour	
	nmpcolour.r = 128
	nmpcolour.g = 128
	nmpcolour.b = 255
	nmpcolour.a = 255
	'	colour difference
	colour_diff = get_colour_distance(avrcolour, nmpcolour)

	#if (_DDSDEBUG = 1)
		cprint using "colour distance = ####.####";colour_diff
		cprint "press any key."
		sleep
	#endif
	
	if (colour_diff < DDS_NORMALMAP_MAXCOLORDISTANCE) then		' threshold value based on the results of the experiment.
		return(true)
	else
		return(false)
	endif
end function

end namespace
