'----------------------------------------------------------
'	DDS File Header structure and flag value definitions
'	and Volition's texture format definition
'	Copyright (c) 2020 Masamichi Kirikae
'----------------------------------------------------------
#include once "crt/stdio.bi"
#include once "crt/string.bi"
#include once "windows.bi"
#include once "dir.bi"
#include once "datatype.bi"
#include once "squish.bi"
#include once "fileio.bi"

' Volition's format value mapping
#define	D3DFMT_DXT1 		400
#define	D3DFMT_DXT3			401	
#define	D3DFMT_DXT5			402
#define	D3DFMT_R5G6B5		403
#define	D3DFMT_A1R5G5B5		404
#define	D3DFMT_A4R4G4B4		405
#define	D3DFMT_R8G8B8		406
#define	D3DFMT_A8R8G8B8		407
#define	D3DFMT_V8U8			408
#define	D3DFMT_CxV8U8		409
#define	D3DFMT_A8			410

' flags for header infomation
#define DDSD_CAPS			&H00000001	' Caps/Caps2 is valid
#define DDSD_HEIGHT			&H00000002	' Height is valid
#define DDSD_WIDTH			&H00000004	' Width is valid
#define DDSD_PITCH			&H00000008	' PitchOrLinearSize is Pitch value
#define DDSD_PIXELFORMAT	&H00001000	' PfSize/PfFlags/RGB... is valid
#define DDSD_MIPMAPCOUNT	&H00020000	' MipMapCount is valid
#define DDSD_LINEARSIZE		&H00080000	' PitchOrLinearSize is LinearSize value
#define DDSD_DEPTH			&H00800000	' Depth is valid

' flags for pixelformat
#define DDPF_ALPHAPIXELS	&H00000001	' Alpha data exists
#define DDPF_ALPHA			&H00000002	' pixel data is only alpha
#define DDPF_FOURCC			&H00000004	' FourCC is valid
#define DDPF_RGB			&H00000040	' Format is defined by RGBBitCount/RBitMask/GBitMask/BBitMask/ABitMask
#define DDPF_LUMINANCE		&H00020000	' Data of 1ch is expanded to all RGB
#define DDPF_BUMPDUDV		&H00080000	' pixel is signed data (for bumpmap)

' flags for Caps1
#define DDSCAPS_ALPHA		&H00000002	' Alpha is included
#define DDSCAPS_COMPLEX		&H00000008	' In case of multiple data (Palette/Mipmap/Cube/Volume etc...)
#define DDSCAPS_TEXTURE		&H00001000	' Always true
#define DDSCAPS_MIPMAP		&H00400000	' Mipmap exists

'flags for Caps2
#define DDSCAPS2_CUBEMAP	&H00000200	' Cubemap exists
#define DDSCAPS2_VOLUME		&H00000200	' data is Volume Texture

type struct_ddsheader
	dwMagic 			as u32	' = 0x20534444
	dwSize 				as u32	' = 124
	dwFlags 			as u32	' valid field flag. contains DDSD_CAPS, DDSD_PIXELFORMAT, DDSD_WIDTH, DDSD_HEIGHT
	dwHeight 			as u32	
	dwWidth				as u32
	dwPitchOrLinear		as u32	' PitchOrLinearSize (pitch:bytes/line LinearSize:bytes/Frame
	dwDepth				as u32	' for volume Texture
	dwMipMapCount		as u32	' store mipmap level
	dwReserved1(10)		as u32	' reserved 11 DWORD (0 to 10) value is 0
	dwPfSize			as u32	' = 32
	dwPfFlags			as u32	' DDPF_RGB / DDPF_FOURCC
	dwFourCC			as u32	' set value when dwPfFlags is DDPF_FOURCC
	dwRGBBitCount		as u32	' bits/pixel with RGB format (must be 0 when DXT1-5)
	dwRBitMask			as u32	' R mask with RGB Format (when DDPF_RGB)
	dwGBitMask			as u32	' G mask with RGB Format (when DDPF_RGB)
	dwBBitMask			as u32	' B mask with RGB Format (when DDPF_RGB)
	dwABitMask			as u32	' alpha mask with RGBA Format (when DDPF_RGB)
	dwCaps1				as u32	' DDSD_CAPS_TEXTURE(must), DDSD_CAPS_MIPMAP | DDSD_CAPS_COMPLEX (size of ddsCaps structure is 16byte)
	dwCaps2				as u32	' for cubemap or volume texutre flag etc.
	Reserved(1)			as u32	' reserved DWORD (0 to 1)  last elemet of ddsCaps structure
	dwReserved2			as u32	' reserved DWORD
end type

#ifndef rgba_data
	type rgba_data
		r as u8
		g as u8
		b as u8
		a as u8
	end type
#endif

#macro rgb_to_rgb565(c, r, g, b)
	' rrrr rggg gggb bbbb
	c = (b) shr 3							' bbbb b000
	c = c or (((g) shr 2) shl 5)		' gggg gg00 
	c = c or (((r) shr 3) shl 11)		' rrrr r000
#endmacro

#macro limits_ubyte(a)
	if (a) > 255 then
		(a) = 255
	elseif (a) < 0 then
		(a) = 0
	endif
#endmacro

#ifndef arrraycopy
	#define arraycopy(dest, src, size) memcpy(@dest(0), @src(0), sizeof((dest(0))) * (size - 1))
#endif

#ifndef integer_sign_inversion
	#define integer_sign_inversion(n) (not n)+1
#endif

#ifndef pi
	const as double pi =	3.14159265358979323846264338327950288
#endif

#ifndef _higher
 #define _higher(a, b, c) if (a > b) then c = a else c = b
#endif

#ifndef _lower
 #define _lower(a, b, c) if (a < b) then c = a else c = b
#endif

#define _DDSDEBUG		0
#define _DDSBENCHMARK	0

#define DDS_GENERIC_ENGINE		1
#define DDS_SR2_ENGINE			4

#define DDS_MIPMAP_DXT1			0
#define DDS_MIPMAP_DXT3_DXT5	1
#define DDS_MIPMAP_NORMALMAP	2

#define DDS_sRGB_COLOUR		1
#define DDS_LINEAR_COLOUR	2

#define DDS_NORMALMAP_MAXCOLORDISTANCE	60.0

type mipmap_generate_director_data
	resize_filter	as zstring*64
	colourspace		as s32
end type

type dds_director_data
	silentmode			as s32
	render_target 		as s32
	colourfit			as s32
	mipmap(0 to 2)	as mipmap_generate_director_data
	elapsed_time		as double
end type

namespace dds

	declare sub init_dds()
	declare function init_ddsheader(byref ddshdr as struct_ddsheader, iwidth as s32, iheight as s32, fmt as s32, mipmapcount as s32) as s32
	declare function load_ddsheader(byref filename as string, byref iwidth as s32, byref iheight as s32, byref pixform as s32) as s32
	declare function load_dds(byref fname as string, image() as u8, byref width as s32, byref height as s32, byref pixform as s32) as s32
	declare function save_dds_rgba(byref fname as string, image_rgba() as u8, iwidth as s32, iheight as s32, pixform as s32, mipmapflag as s32) as s32
	declare function check_dxtc_sourceimagesize(iwidth as s32, iheight as s32, mip as s32) as s32
	declare function get_maxmipmaplevel(iwidth as s32, iheight as s32) as s32

	'declare sub convert_RGB_to_YCbCr(byref r as double, byref g as double, byref b as double, byref Y as double, byref Cb as double, byref Cr as double)
	'declare sub convert_YCbCr_to_RGB(byref r as double, byref g as double, byref b as double, byref Y as double, byref Cb as double, byref Cr as double)
	declare sub normalmap_to_tgn(normal_rgba() as u8, tgn_rgba() as u8, iwidth as s32, iheight as s32)
	declare function tgn_to_normalmap(tgn_rgba() as u8, normal_rgba() as u8, iwidth as s32, iheight as s32) as s32
	declare function get_colour_distance(c1 as rgba_data, c2 as rgba_data) as double option ("sse")
	'declare function get_average_colour_fast(image_rgba() as u8, iwidth as s32, iheight as s32) as rgba_data
	declare function get_average_colour(image_rgba() as u8, iwidth as s32, iheight as s32) as rgba_data
	declare sub rgb565_to_rgba(byref c as u16, byref r as u8, byref g as u8, byref b as u8, byref a as u8)
	declare function is_normalmap(image_rgba() as u8, iwidth as s32, iheight as s32, pixform as s32) as s32
	declare function get_psnr(ref_rgba() as u8, cmp_rgba() as u8, iwidth as s32, iheight as s32) as double
	declare function get_suitable_format(image_rgba() as u8, iwidth as s32, iheight as s32) as s32

	'declare function fast_round(byval x as double) as long
	'declare function fast_sin(byval x as double) as double option ("SSE")
	'declare function sRGB_to_linear(c as double) as double
	'declare function Linear_to_sRGB(c as double) as double
	
	common shared dds_director as dds_director_data

end namespace
