'-------------------------------------------------
'		extract/rebuild peg_pc, g_peg_pc file
'		Copyright (c) 2019 Masamichi Kirikae
'-------------------------------------------------

#include once "pegfile.bi"

namespace peg_pc

'--------------------------------------------------------------------
'	initialize peg library director variables
'--------------------------------------------------------------------
sub init_peg()
	peg_director.check_imagesize_difference = true
	peg_director.suppress_extracted = false	
end sub

'--------------------------------------------------------------------
'	extract <name> from fullpath\<name>.peg_pc
'--------------------------------------------------------------------
function get_pegname(byref fname as string) as string
	dim i as s32, lg as s32, st as s32, ed as s32
	dim pegname as string
	' search "\" from right
	lg = len(fname)
	st = 0
	for i = lg to 1 step -1
		if (mid(fname, i, 1) = chr(&H5C)) then
			st = i + 1
			exit for
		endif
	next
	if (st = 0) then return("")	' in case "\" not found
	' search "." from right
	ed = 0
	for i = lg to 1 step -1
		if (mid(fname, i, 1) = ".") then
			ed = i - 1
			exit for
		endif
	next
	if (ed = 0) then return("")	' in case "." not found
	pegname = mid(fname, st, ed - st + 1)
	return(pegname)
end function

'--------------------------------------------------------------------
'	save_dds_raw() Makes dds file from raw compressed data
'	return value mipmapcount:success 0:error
'--------------------------------------------------------------------
function save_dds_raw(byref fname as string, imgw as s32, imgh as s32, rawdata() as u8, rawsize as s32, pixform as s32, mipmapcount as s32) as s32

	const attrib_mask = fbArchive
	dim ddshdr as struct_ddsheader
	dim as s32 fh, ret
	
	if ((imgw = 0) or (imgh = 0)) then return(0)	' Error

	ret = dds.init_ddsheader(ddshdr, imgw, imgh, pixform, mipmapcount)
	if (ret = 0) then return(0)	' Error. initialize failed.

	' delete file when <fname> is already exists
	if (dir(fname, attrib_mask) <> "") then
		kill(fname)
	endif
	
	fh = freefile
	open fname for output as #fh

	put #fh, ,ddshdr
	put #fh, ,rawdata(0), rawsize	
	close #fh

	return(mipmapcount)
end function

'--------------------------------------------------------------------
'	write peg header
'--------------------------------------------------------------------
sub write_pegheader(fh as s32, byref peg as peg_header)
	put #fh, , peg.magic				'string*4
	put #fh, , peg.version			'u16
	put #fh, , peg.unknown06		'u16
	put #fh, , peg.filesize			'u32
	put #fh, , peg.datafilesize	'u32
	put #fh, , peg.entrycount		'u16
	put #fh, , peg.unknown12		'u16
	put #fh, , peg.framecount		'u16
	put #fh, , peg.unknown16		'u16
end sub

'--------------------------------------------------------------------
'	read peg frame data (48bytes)
'--------------------------------------------------------------------
sub read_pegframe(fh as s32, byref frame as peg_frame)
	get #fh, , frame.offset
	get #fh, , frame.width
	get #fh, , frame.height
	get #fh, , frame.format
	get #fh, , frame.unknown0A
	get #fh, , frame.unknown0C
	get #fh, , frame.frames
	get #fh, , frame.unknown12
	get #fh, , frame.unknown16
	get #fh, , frame.unknownflags
	get #fh, , frame.mipmapcount
	get #fh, , frame.size
	get #fh, , frame.unknown20
	get #fh, , frame.unknown24
	get #fh, , frame.unknown28
	get #fh, , frame.unknown2C
end sub

'--------------------------------------------------------------------
'	write peg frame data (48bytes)
'--------------------------------------------------------------------
sub write_pegframe(fh as s32, byref frame as peg_frame)
	put #fh, , frame.offset
	put #fh, , frame.width
	put #fh, , frame.height
	put #fh, , frame.format
	put #fh, , frame.unknown0A
	put #fh, , frame.unknown0C
	put #fh, , frame.frames
	put #fh, , frame.unknown12
	put #fh, , frame.unknown16
	put #fh, , frame.unknownflags
	put #fh, , frame.mipmapcount
	put #fh, , frame.size
	put #fh, , frame.unknown20
	put #fh, , frame.unknown24
	put #fh, , frame.unknown28
	put #fh, , frame.unknown2C
end sub

'--------------------------------------------------------------------
'	this function used for realtime texture view
'--------------------------------------------------------------------
function get_pegdata(byref filename as string, byref peg as peg_header, names() as string, frames() as peg_frame) as s32
	dim as string  peg_file
	dim as s32 peg_fh
	dim as s32 i, j, n, x, y
	dim c as u8, crgba as rgba_data

	peg_file = libfio.change_filename_ext(filename, ".peg_pc")
		
	peg_fh = freefile
	open peg_file for binary access read as #peg_fh
	
	peg.magic = libfio.fgetstring(peg_fh, 4)
	if (peg.magic <> chr(&h47, &H45, &H4B, &H56)) then
		cprint "ERROR:Header must be GEKV"
		close #peg_fh
		return(F_ERROR)
	endif

	get #peg_fh, , peg.version
	if (peg.version <> 10) then
		cprint "ERROR:Unsupported version file"
		close #peg_fh
		return(F_ERROR)
	endif

	get #peg_fh, , peg.unknown06
	get #peg_fh, , peg.filesize
	get #peg_fh, , peg.datafilesize
	get #peg_fh, , peg.entrycount
	get #peg_fh, , peg.unknown12
	get #peg_fh, , peg.framecount
	get #peg_fh, , peg.unknown16

	redim names(peg.entrycount) as string	
	fpseek(peg_fh, &H18 + (&h30 * peg.framecount))
	for i = 0 to peg.entrycount - 1
		names(i) = libfio.fgetstringz(peg_fh)
	next

	' read entries
	redim frames(peg.entrycount) as peg_frame
	dim as s32 totalflames, compress_size, decompress_size, imagew, imageh, flag
	
	fpseek(peg_fh, &h18)
	for i = 0 to peg.entrycount - 1
		 read_pegframe(peg_fh, frames(i))
	next

	close #peg_fh

	return(F_NOERROR)

end function

'--------------------------------------------------------------------
'	this function used for realtime texture view
'--------------------------------------------------------------------
function get_image_gpeg(byref gpeg_name as string, image_rgba() as u8, peg as peg_header, frames() as peg_frame, entry as s32) as s32
	dim as s32 gpeg_fh, imagew, imageh, decompress_size, compress_size
	dim as s32 i, j, n, x, y, flag
	dim c as u8, c16 as u16, crgba as rgba_data

	gpeg_fh = freefile
	open gpeg_name for binary access read as #gpeg_fh

	imagew = cast(s32, frames(entry).width)
	imageh = cast(s32, frames(entry).height)
	decompress_size = imagew * imageh * 4
	
	' display texuture
	select case frames(entry).format
		case D3DFMT_DXT1
			flag = squish_kDxt1
			compress_size = squish.GetStorageRequirements(cint(imagew), cint(imageh), flag)
			'	read compressed data
			fpseek(gpeg_fh, frames(entry).offset)
			get #gpeg_fh, ,squish.dxtblocks(0) ,compress_size
			'	decompress
			squish.DecompressImage(@squish.image_rgba(0), cint(imagew), cint(imageh), @squish.dxtblocks(0), flag)
		case D3DFMT_DXT3
			flag = squish_kDxt3
			compress_size = squish.GetStorageRequirements(cint(imagew), cint(imageh), flag)
			'	read compressed data
			fpseek(gpeg_fh, frames(entry).offset)
			get #gpeg_fh, ,squish.dxtblocks(0) ,compress_size
			'	decompress
			squish.DecompressImage(@squish.image_rgba(0), cint(imagew), cint(imageh), @squish.dxtblocks(0), flag)
		case D3DFMT_DXT5
			flag = squish_kDxt5
			compress_size = squish.GetStorageRequirements(cint(imagew), cint(imageh), flag)
			'	read compressed data
			fpseek(gpeg_fh, frames(entry).offset)
			get #gpeg_fh, ,squish.dxtblocks(0) ,compress_size
			'	decompress
			squish.DecompressImage(@squish.image_rgba(0), cint(imagew), cint(imageh), @squish.dxtblocks(0), flag)
		case D3DFMT_A8R8G8B8
			'	read uncompressed argb data (not rgba)
			fpseek(gpeg_fh, frames(entry).offset)
			n = 0
			for y = 0 to imageh - 1
				for x = 0 to imagew - 1
					get #gpeg_fh, ,crgba.b	'bgra (little endian)
					get #gpeg_fh, ,crgba.g
					get #gpeg_fh, ,crgba.r
					get #gpeg_fh, ,crgba.a
					'	argb -> rgba
					squish.image_rgba(n) = crgba.r: n+=1
					squish.image_rgba(n) = crgba.g: n+=1
					squish.image_rgba(n) = crgba.b: n+=1
					squish.image_rgba(n) = crgba.a: n+=1
				next
			next
		case D3DFMT_R5G6B5
			'	read uncompressed r5g6b5 data
			fpseek(gpeg_fh, frames(entry).offset)
			n = 0
			for y = 0 to imageh - 1
				for x = 0 to imagew - 1
					get #gpeg_fh, ,c16	' u16 R5G6B5
					dds.rgb565_to_rgba(c16, crgba.r, crgba.g, crgba.b, crgba.a)	' always alpha = &HFF 
					squish.image_rgba(n) = crgba.r: n+=1
					squish.image_rgba(n) = crgba.g: n+=1
					squish.image_rgba(n) = crgba.b: n+=1
					squish.image_rgba(n) = crgba.a: n+=1
				next
			next
		case else
	end select

	close #gpeg_fh

	redim image_rgba(decompress_size - 1) as u8
	arraycopy(image_rgba, squish.image_rgba, decompress_size)

	return(F_NOERROR)
		
end function

'--------------------------------------------------------------------
'	extract textures from peg_pc and g_peg_pc
'--------------------------------------------------------------------
function extract_peg(byref filename as string, destpath as string) as s32
	dim as string pegname, desc_file, dds_file, extract_folder, gpeg_file, png_file
	dim as s32 peg_fh, gpeg_fh
	dim peg as peg_header, ddshdr as struct_ddsheader
	dim as s32 i, j, n, x, y, max_mip, ret, attrib_mask
	dim as u32 offset_adr
	dim c as u8, c16 as u16, crgba as rgba_data
	dim as s32 compress_size, image_size, imagew, imageh, flag, framecount, frame_max
	dim image_rgba(any) as u8
	dim rawdata(any) as u8

	pegname = get_pegname(filename)
	gpeg_file = libfio.change_filename_ext(filename, ".g_peg_pc")

	' setup extract folder
	if (destpath = "") then
		if (peg_director.suppress_extracted) then
			extract_folder = libfio.change_filename_ext(filename, "")
		else
			extract_folder = libfio.change_filename_ext(filename, "_extracted")
		endif
	else
		extract_folder = destpath
		n = len(extract_folder)
		if (right(extract_folder, 1) = "\") then
			extract_folder = left(extract_folder, n - 1)
		endif
	endif

	if (dir(extract_folder, fbDirectory) = "") then mkdir(extract_folder)
	
	desc_file = extract_folder + "\" + pegname + ".peg_desc"
	
	peg_fh = freefile
	open filename for binary access read as #peg_fh
	if (err > 1) then
		cprint "ERROR:Can't open peg_pc file"
		close #peg_fh
		return(F_ERROR)
	endif

	xml.new_doc( desc_file, "PegDescription" )
	
	peg.magic = libfio.fgetstring(peg_fh, 4)
	if (peg.magic <> chr(&h47, &H45, &H4B, &H56)) then
		cprint "ERROR:Header must be GEKV"
		close #peg_fh
		return(F_ERROR)
	endif

	get #peg_fh, , peg.version
	if (peg.version <> 10) then
		cprint "ERROR:Unsupported version file"
		close #peg_fh
		return(F_ERROR)
	endif

	get #peg_fh, , peg.unknown06
	get #peg_fh, , peg.filesize
	get #peg_fh, , peg.datafilesize
	get #peg_fh, , peg.entrycount
	get #peg_fh, , peg.unknown12
	get #peg_fh, , peg.framecount
	get #peg_fh, , peg.unknown16
	
	xml.write_single_element( "BigEndian",			"False"						)
	xml.write_single_element( "Magic",				"GEKV"						)
	xml.write_single_element( "Version",			"10"							)
	xml.write_single_element( "Unknown06",			hex(peg.unknown06, 4)	)
	xml.write_single_element( "FileSize",			str(peg.filesize)			)
	xml.write_single_element( "DataFileSize",		str(peg.datafilesize)	)
	xml.write_single_element( "EntryCount",		str(peg.entrycount)		)
	xml.write_single_element( "Unknown12",			hex(peg.unknown12, 4)	)
	xml.write_single_element( "FrameCount",		str(peg.framecount)		)
	xml.write_single_element( "Unknown16",			hex(peg.unknown16, 4)	)

	dim names(peg.entrycount) as string
	fpseek(peg_fh, &H18 + (&h30 * peg.framecount))
	for i = 0 to peg.entrycount - 1
		names(i) = libfio.fgetstringz(peg_fh)
	next

	xml.new_element( "Entries" )

	gpeg_fh = freefile
	open gpeg_file for binary access read as #gpeg_fh
	
	' read entries/frames
	dim frame(peg.framecount) as peg_frame
	dim formatname as string
	
	framecount = 0
	fpseek(peg_fh, &h18)
	for i = 0 to peg.entrycount - 1
		
		xml.new_element( "Entry" )							'	<Entry>
		xml.write_single_element( "Name", names(i) )	'	<Name>***</Name>
		xml.new_element( "Frames" )						'	<Frames>

		read_pegframe(peg_fh, frame(framecount))
		
		frame_max = frame(framecount).frames
		if (frame_max > 1) then
			xml.write_single_element("FrameCount", str(frame_max))
		endif
		
		for j = 0 to frame_max - 1
			if (j > 0) then
				framecount += 1
				read_pegframe(peg_fh, frame(framecount))
			end if

			if (not(dds.dds_director.silentmode)) then
				cprint names(i); "_"; str(j)
			endif

			' one frame data
			xml.new_element( "Frame" )		'	<Frame>
	
			xml.write_single_element( "Offset", 	str(frame(framecount).offset)	)
			xml.write_single_element( "Width", 		str(frame(framecount).width)	)
			xml.write_single_element( "Height", 	str(frame(framecount).height)	)
	
			imagew = cast(s32, frame(framecount).width)
			imageh = cast(s32, frame(framecount).height)

			image_size = imagew * imageh * 4
			redim image_rgba(image_size - 1) as u8
			
			'	write format to xml
			select case frame(framecount).format
				case D3DFMT_DXT1
					xml.write_single_element( "Format", "DXT1" )
					flag = squish_kDxt1
					compress_size = squish.GetStorageRequirements(imagew, imageh, flag)
				case D3DFMT_DXT3
					xml.write_single_element( "Format", "DXT3" )
					flag = squish_kDxt3
					compress_size = squish.GetStorageRequirements(imagew, imageh, flag)
				case D3DFMT_DXT5
					xml.write_single_element( "Format", "DXT5" )
					flag = squish_kDxt5
					compress_size = squish.GetStorageRequirements(imagew, imageh, flag)
				case D3DFMT_A8R8G8B8
					xml.write_single_element( "Format", "A8R8G8B8" )
				case D3DFMT_R5G6B5
					xml.write_single_element( "Format", "R5G6B5" )
				case else	'	unknown format
					xml.write_single_element( "Format", str(frame(i).format) )
			end select

			'	make png file(s)
			png_file = extract_folder + "\" + names(i) + "_" + str(j) + ".png"
			offset_adr = frame(framecount).offset
			fpseek(gpeg_fh, offset_adr)
			select case frame(framecount).format
				case D3DFMT_DXT1, D3DFMT_DXT3, D3DFMT_DXT5
					'	read compressed data
					get #gpeg_fh, ,squish.dxtblocks(0) ,compress_size
					'	decompress
					squish.DecompressImage(@squish.image_rgba(0), imagew, imageh, @squish.dxtblocks(0), flag)
					'	save image to png
					libpng.save_png(png_file, squish.image_rgba(), imagew, imageh, 32)
				case D3DFMT_A8R8G8B8
					'	read uncompressed argb data (not rgba)
					n = 0
					for y = 0 to imageh - 1
						for x = 0 to imagew - 1
							get #gpeg_fh, ,crgba.b	'bgra (little endian)
							get #gpeg_fh, ,crgba.g
							get #gpeg_fh, ,crgba.r
							get #gpeg_fh, ,crgba.a
							'	argb -> rgba
							image_rgba(n) = crgba.r: n+=1
							image_rgba(n) = crgba.g: n+=1
							image_rgba(n) = crgba.b: n+=1
							image_rgba(n) = crgba.a: n+=1
						next
					next
					'	save image to png
					libpng.save_png(png_file, image_rgba(), imagew, imageh, 32)
				case D3DFMT_R5G6B5
					n = 0
					for y = 0 to imageh - 1
						for x = 0 to imagew - 1
							get #gpeg_fh, ,c16	'u16 rrrrrggg gggbbbbb
							dds.rgb565_to_rgba(c16, crgba.r, crgba.g, crgba.b, crgba.a) 
							image_rgba(n) = crgba.r: n+=1
							image_rgba(n) = crgba.g: n+=1
							image_rgba(n) = crgba.b: n+=1
							image_rgba(n) = crgba.a: n+=1
						next
					next
					'	save image to png
					libpng.save_png(png_file, image_rgba(), imagew, imageh, 32)
				case else
			end select

			'	Make dds File(s)		
			redim rawdata(frame(framecount).size - 1) as u8
			offset_adr = frame(framecount).offset
			fpseek(gpeg_fh, offset_adr)
			dds_file = extract_folder + "\" + names(i) + "_" + str(j) + ".dds"
			get #gpeg_fh, , rawdata(0), frame(framecount).size
			'	write ddsheader + rawdata. 
			ret = save_dds_raw(dds_file, imagew, imageh, rawdata(), frame(framecount).size, frame(framecount).format, frame(framecount).mipmapcount)
			if (ret <> frame(framecount).mipmapcount) then
				cprint "Error:unknown error in save_dds_raw() function."
				xml.close_doc()
				close #peg_fh
				close #gpeg_fh
				return(F_ERROR)
			endif	

			select case frame(framecount).format
				case D3DFMT_DXT1, D3DFMT_DXT3, D3DFMT_DXT5
					max_mip = dds.get_maxmipmaplevel(imagew, imageh)
					'	check mipmaplevel level restriction
					if (frame(framecount).mipmapcount < max_mip) then
						frame(framecount).mipmapcount += 100
					endif
				case else
			end select

			xml.write_single_element( "Unknown0A",			hex(frame(framecount).unknown0A, 4)		)
			xml.write_single_element( "Unknown0C",			hex(frame(framecount).unknown0C, 8)		)
			xml.write_single_element( "Frames",				str(frame(framecount).frames)          )
			xml.write_single_element( "Unknown12",			hex(frame(framecount).unknown12, 8)		)
			xml.write_single_element( "Unknown16",			hex(frame(framecount).unknown16, 8)		)
			xml.write_single_element( "UnknownFlags1A",	hex(frame(framecount).unknownflags, 2)	)
			xml.write_single_element( "MipmapCount",		str(frame(framecount).mipmapcount)		)
			xml.write_single_element( "Size",				str(frame(framecount).size)				)
			xml.write_single_element( "Unknown20",			hex(frame(framecount).unknown20, 8)		)
			xml.write_single_element( "Unknown24",			hex(frame(framecount).unknown24, 8)		)
			xml.write_single_element( "Unknown28",			hex(frame(framecount).unknown28, 8)		)
			xml.write_single_element( "Unknown2C",			hex(frame(framecount).unknown2C, 8)		)
		
			xml.close_element()	'	</Frame>
		next
		xml.close_element()	'	</Frames>
		xml.close_element()	'	</Entry>
		framecount += 1
	next

	xml.close_element()
	xml.close_doc()
	
	close #peg_fh
	close #gpeg_fh

	if (not(dds.dds_director.silentmode)) then
		cprint ""
		cprint "extract "; peg.entrycount; " entries / ";peg.framecount; " frames."
		cprint ""
	endif
	
	return(F_NOERROR)

end function

'--------------------------------------------------------------------
'	rebuid gpeg_pc and peg_pc files
'--------------------------------------------------------------------
function rebuild_peg(byref filename as string, byref destpath as string) as s32
	dim image_rgba(any) as u8
	dim as s32 i, j, n, attrib_mask, result
	dim as s32 peg_fh, gpeg_fh, dds_fh
	dim as s32 png_imgw, png_imgh, bytepp, dds_imgw, dds_imgh, pixform, makeddsflag
	dim as u32 mip, actual_mip, mip_restricted, size_raw, offset_address, byteborder
	dim as s32 framecount, frame_max, compress_size, decompress_size, imagew, imageh, flag
	dim as u8 c, fill_zero(14) = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
	dim as string pegname, peg_file, gpeg_file, png_file, dds_file, source_folder, rebuild_folder, ename, value, formatname, cmdline
	dim peg as peg_header
	dim as double pngftime, ddsftime
	dim as zstring * 256 path_zstr  = ""

	dim names(any) as string
	dim framenames(any) as string
	dim frame(any) as peg_frame

	pegname = get_pegname(filename)
	source_folder = libfio.get_directory(filename)

	' setup rebuild folder
	if (destpath = "") then
		rebuild_folder = libfio.change_filename_ext(filename, "_rebuild")
	else
		rebuild_folder = destpath
		n = len(rebuild_folder)
		if (right(rebuild_folder, 1) = "\") then
			rebuild_folder = left(rebuild_folder, n - 1)
		endif
	endif
	
	peg_file = rebuild_folder + "\" + pegname + ".peg_pc"
	gpeg_file = rebuild_folder + "\" + pegname + ".g_peg_pc"
	
	xml.read_doc(filename)
	xml.read_element(ename, value)	' <PegDescription>
	if (ename <> "PegDescription") then
		cprint "ERROR:Not peg_desc file or syntax error."
		xml.close_doc()
		return(F_ERROR)
	endif
	result = xml.read_element(ename, value)	' BigEndian
	if (value <> "False") then
		cprint "ERROR:Endian not match or syntax error."
		xml.close_doc()
		return(F_ERROR)
	endif
	if (value <> "False") then
		cprint "ERROR:Not support bigendian byteorder or syntax error."
		xml.close_doc()
		return(F_ERROR)
	endif

	'	read peg header
	peg.magic = chr(&h47, &H45, &H4B, &H56)	'	GEKV
	for i = 0 to 8
		xml.read_element(ename, value)
		select case ename
			case "Version"
				peg.version = val(value)
			case "Unknown06"
				peg.unknown06 = val("&H" + value)
			case "FileSize"
				peg.filesize = val(value)
			case "DataFileSize"
				peg.datafilesize = val(value)
			case "EntryCount"
				peg.entrycount = val(value)
			case "Unknown12"
				peg.unknown12 = val("&H" + value)
			case "FrameCount"
				peg.framecount = val(value)
			case "Unknown16"
				peg.unknown16 = val("&H" + value)
			case else
		end select
	next

	'-----------------------------------------------------------------
	' read entries data
	'-----------------------------------------------------------------
	redim names(peg.entrycount - 1) as string
	redim framenames(peg.framecount - 1) as string
	redim frame(peg.framecount - 1) as peg_frame

	result = xml.read_element(ename, value)	' <Entries>
	framecount = 0

	if (not(dds.dds_director.silentmode)) then
		cprint using "entrycount = ####"; peg.entrycount
	endif
	
	for i = 0 to peg.entrycount - 1

		xml.read_element(ename, value)	' <Entry>
		xml.read_element(ename, value)	' <Name>
		names(i) = value
		xml.read_element(ename, value)	' <Frames>
		xml.read_element(ename, value)	' <Frame> or <FrameCount>
		select case ename			
			case "FrameCount"
				frame_max = val(value)
				xml.read_element(ename, value)	' <Frame>
			case "Frame"
				frame_max = 1
			case else
				cprint "Error:XML Syntax Error in "; names(i)
				xml.close_doc()
				return(F_ERROR)	
		end select

		'-----------------------------------------------------------------
		'	frames
		'-----------------------------------------------------------------
		for j = 0 to frame_max - 1
			if (j > 0) then
				xml.read_element(ename, value)	' <Frame>				
			endif
			framenames(framecount) = names(i) + "_" + str(j)
			xml.read_element(ename, value)	' Offset
			frame(framecount).offset = val(value)
			xml.read_element(ename, value)	' Width
			frame(framecount).width = val(value)
			xml.read_element(ename, value)	' Height
			frame(framecount).height = val(value)
			xml.read_element(ename, value)	' Format
			select case value
				case "DXT1"
					frame(framecount).format = D3DFMT_DXT1
				case "DXT3"
					frame(framecount).format = D3DFMT_DXT3
				case "DXT5"
					frame(framecount).format = D3DFMT_DXT5
				case "A8R8G8B8"
					frame(framecount).format = D3DFMT_A8R8G8B8
				case "R5G6B5"
					frame(framecount).format = D3DFMT_R5G6B5				
				case else
					xml.close_doc()
					cprint using "Error:unsupported image format:& in &"; value; framenames(framecount)
					return(F_ERROR)
			end select
			xml.read_element(ename, value)			' Unknown0A
			frame(framecount).unknown0A = val("&H" + value)
			xml.read_element(ename, value)			' Unknown0C
			frame(framecount).unknown0C = val("&H" + value)		
			xml.read_element(ename, value)			' Frames
			frame(framecount).frames = val(value)
			
			xml.read_element(ename, value)			' Unknown12
			frame(framecount).unknown12 = val("&H" + value)
			xml.read_element(ename, value)			' Unknown16
			frame(framecount).unknown16 = val("&H" + value)
			xml.read_element(ename, value)			' UnknownFlags1A
			frame(framecount).unknownflags = val("&H" + value)
			xml.read_element(ename, value)			' MipMapCount
			frame(framecount).mipmapcount = val(value)
			xml.read_element(ename, value)			' Size
			frame(framecount).size = val(value)
			xml.read_element(ename, value)			' Unknown20
			frame(framecount).unknown20 = val("&H" + value)
			xml.read_element(ename, value)			' Unknown24
			frame(framecount).unknown24 = val("&H" + value)
			xml.read_element(ename, value)			' Unknown28
			frame(framecount).unknown28 = val("&H" + value)
			xml.read_element(ename, value)			' Unknown2C
			frame(framecount).unknown2C = val("&H" + value)
			xml.read_element(ename, value)			' </Frame>
			framecount += 1
		next
		xml.read_element(ename, value)			' </Flames>
		xml.read_element(ename, value)			' </Entry>
	next

	xml.read_element(ename, value)			' </PegDescription>
	
	result = xml.close_doc()	'  no need to check return value
	
	'-----------------------------------------------------------------
	'	make folder
	'-----------------------------------------------------------------
	if (dir(rebuild_folder, fbDirectory) = "") then mkdir(rebuild_folder)
	
	'-----------------------------------------------------------------
	'	make g_peg_pc file
	'-----------------------------------------------------------------
	gpeg_fh = freefile
	open gpeg_file for output as #gpeg_fh

	attrib_mask = fbArchive
	offset_address = 0
	for i = 0 to peg.framecount - 1

		if (not(dds.dds_director.silentmode)) then
			cprint "processing "; framenames(i)
		endif
		
		frame(i).offset = offset_address
		png_file = source_folder + framenames(i) + ".png"
		dds_file = source_folder + framenames(i) + ".dds"
		' file existence check
		if (dir(png_file, attrib_mask) = "") then
			cprint using "Error:& not found."; libfio.get_filename_only(png_file)
			close gpeg_fh
			return(F_ERROR)
		endif

		libpng.load_png(png_file, image_rgba(), png_imgw, png_imgh, bytepp)

		'-----------------------------------------------------------------
		'  PNG image is converted to DDS when
		'  1. dds file not exist.
		'  2. there's a difference between png imagesize and peg_desc imagesize.
		'	3. there's a difference between png imagesize and dds imagesize.
		'	4. there's a difference between dds format and peg_desc format.
		'  5. png filetime newer than dds filetime.
		'	6. there's a difference between dds mipmap and peg_desc mipmap when frame(i)mipmaplevel > 100
		'-----------------------------------------------------------------

		if (frame(i).mipmapcount > 100) then
			mip_restricted = true
			actual_mip = frame(i).mipmapcount - 100
		else
			mip_restricted = false
			actual_mip = frame(i).mipmapcount
		endif
		
		'-----------------------------------------------------------------
		'	note: check_imagesize_difference = false in sr2chunktool.
		'	compare png imagesize and described imagesize 
		'-----------------------------------------------------------------
		if (peg_director.check_imagesize_difference) then
			if ((png_imgw <> frame(i).width) or (png_imgh <> frame(i).height)) then
				makeddsflag = true
			else
				makeddsflag = false
			endif
		else
			makeddsflag = false
		endif
		'	check dds file
		if (makeddsflag = false) then
			if (dir(dds_file, attrib_mask) = "") then		'	dds file not exist.
				makeddsflag = true
			else
				'	get dds file infomation
				mip = dds.load_ddsheader(dds_file, dds_imgw, dds_imgh, pixform)
				if (mip = 0) then
					cprint using "Error:fail to load &."; libfio.get_filename_only(dds_file)
					close #gpeg_fh
					return(F_ERROR)
				endif
				
				'	get png and dds filetime
				pngftime = filedatetime(png_file)
				ddsftime = filedatetime(dds_file)

				'	check dds file infomation
				if ((png_imgw <> dds_imgw) or (png_imgh <> dds_imgh)) then		'	different image size
					makeddsflag = true
				elseif (pixform <> cast(s32, frame(i).format)) then			'	different image format
					makeddsflag = true
				elseif (pngftime > ddsftime) then									'	png file newer than dds file
					makeddsflag = true
				elseif (mip_restricted and (mip <> actual_mip)) then			'	different mipmaplevel (when restricted)
					makeddsflag = true
				else
					'	use dds file to build
					'	not need to convert png to dds
					frame(i).width = cast(u16, dds_imgw)	'	png_imgw = dds_imgw
					frame(i).height = cast(u16, dds_imgh)	'	png_imgh = dds_imgh
					frame(i).mipmapcount = cast(u8, mip)	'	set actual mipmapcount
				endif
			endif
		endif
			
		if (makeddsflag) then		' convert PNG image to DDS
			' ignore image width and height in .peg_desc file.
			frame(i).width = cast(u16, png_imgw)
			frame(i).height = cast(u16, png_imgh)
			pixform = cast(s32, frame(i).format)
			mip = cast(s32, frame(i).mipmapcount)
			'-----------------------------------------------------------------
			'	MipmapCount that described in .peg_desc file
			'	when MipmapCount is 1 to 99 mipmap level is automatically set
			'	when MipmapCount is 0 mipmap is not generated (mipmap level = 1)
			'	when MipmapCount is > 100 mipmap level is restricted to (MipmapCount - 100)
			'-----------------------------------------------------------------
			n = dds.save_dds_rgba(dds_file, image_rgba(), png_imgw, png_imgh, pixform, mip)
			
			frame(i).mipmapcount = cast(u8, n)
		else
			if (mip_restricted) then frame(i).mipmapcount = cast(u8, actual_mip)	' set actual mipmapcount when restricted
		end if
		
		'	get raw data in DDS file
		size_raw = filelen(dds_file) - 128
		redim buf(size_raw - 1) as u8
		
		dds_fh = freefile
		open dds_file for binary access read as #dds_fh
		fpseek(dds_fh, 128)
		get #dds_fh, , buf(0), size_raw
		close #dds_fh
		
		put #gpeg_fh, , buf(0), size_raw

		frame(i).size = cast(u32, size_raw)
	
		offset_address += size_raw
		'	data offset value must be a multiple of 16.
		byteborder = offset_address mod 16
		' 0 fill 16byte border area
		if (byteborder) then
			put #gpeg_fh, ,fill_zero(0), byteborder
			offset_address += byteborder
		end if
	next
	
	close #gpeg_fh
	
	'-----------------------------------------------------------------
	'	make peg_pc file
	'-----------------------------------------------------------------
	peg.datafilesize = filelen(gpeg_file)
	
	peg.filesize = &H18 + (&h30 * peg.framecount)	' &H18 is header structure size. &H30 is frame structure size.
	for i = 0 to peg.entrycount - 1
		peg.filesize += len(names(i)) + 1	'	+1 means chr(0) (stringz)
	next

	peg_fh = freefile
	open peg_file for output as #peg_fh
	'	header
	
	write_pegheader(peg_fh, peg)
	'	entries
	for i = 0 to peg.framecount - 1
		write_pegframe(peg_fh, frame(i))
	next
	'	entrynames (as stringz)	
	for i = 0 to peg.entrycount - 1
		put #peg_fh, ,names(i)
		put #peg_fh, ,chr(0)
	next
	
	close #peg_fh

	if (not(dds.dds_director.silentmode)) then
		cprint ""
		cprint "total "; peg.entrycount; " entries / "; peg.framecount; " frames."
		cprint ""
	end if
	
	return(F_NOERROR)
		
end function

end namespace