'--------------------------------------------------------------------
'	libsquish 1.15
'	DXT1, DXT3, DXT5, BC4, BC5 Compression and Decompression for fbc
'	Copyright (c) 2019 Masamichi Kirikae
'	Original Copyright (c) 2006 Simon Brown
'--------------------------------------------------------------------
#include once "datatype.bi"

#ifndef SQUISH_H
#define SQUISH_H
#endif

#define squish_kDxt1		(1 shl 0)	' Use DXT1 compression.
#define squish_kDxt3		(1 shl 1)	' Use DXT3 compression
#define squish_kDxt5		(1 shl 2)	' Use DXT5 compression.
#define squish_kBc4			(1 shl 3)	' Use BC4 compression.
#define squish_kBc5			(1 shl 4)	' Use BC5 compression.

#define squish_kColourClusterFit			(1 shl 5)	' Use a slow but high quality colour compressor (the default).
#define squish_kColourRangeFit				(1 shl 6)	' Use a fast but low quality colour compressor.
#define squish_kWeightColourByAlpha			(1 shl 7)	' Weight the colour by alpha during cluster fit (disabled by default).
#define squish_kColourIterativeClusterFit	(1 shl 8)	' Use a very slow but very high quality colour compressor.
#define squish_kSourceBGRA					(1 shl 9)	' Source is BGRA rather than RGBA.

extern "c++" lib "squish"
	namespace squish alias "squish"

	enum squishflags
		kDxt1						=	(1 shl 0)	' Use DXT1 compression.
		kDxt3						=	(1 shl 1)	' Use DXT3 compression
		kDxt5						=	(1 shl 2)	' Use DXT5 compression.
		kBc4						=	(1 shl 3)	' Use BC4 compression.
		kBc5						=	(1 shl 4)	' Use BC5 compression.
		kColourClusterFit			=	(1 shl 5)	' Use a slow but high quality colour compressor (the default).
		kColourRangeFit				=	(1 shl 6)	' Use a fast but low quality colour compressor.
		kWeightColourByAlpha		=	(1 shl 7)	' Weight the colour by alpha during cluster fit (disabled by default).
		kColourIterativeClusterFit	=	(1 shl 8)	' Use a very slow but very high quality colour compressor.
		kSourceBGRA					=	(1 shl 9)	' Source is BGRA rather than RGBA.
	end enum
		
	declare sub CompressMasked(byval rgba as const u8 ptr, byval mask as s32, byval blocks as any ptr, byval flags as s32, byval metric as float ptr)
	#macro SqCompress(rgba, block, flags, metric)
		CompressMasked(rgba, &HFFFF, block, flags, metric)
	#EndMacro
	declare sub Decompress(byval rgba as u8 ptr, byval blocks as const any ptr, byval flags as s32)
	declare function GetStorageRequirements(byval width as s32, byval height as s32, byval flags as s32) as s32
	declare sub CompressImage overload (byval rgba as const u8 ptr, byval width as s32, byval height as s32, byval pitch as s32, byval blocks as any ptr, byval flags as s32, byval metric as float ptr)
	declare sub CompressImage overload (byval rgba as const u8 ptr, byval width as s32, byval height as s32, byval blocks as any ptr, byval flags as s32, byval metric as float ptr)
	declare sub DecompressImage overload (byval rgba as u8 ptr, byval width as s32, byval height as s32, byval pitch as s32, byval blocks as const any ptr, byval flags as s32)
	declare sub DecompressImage overload (byval rgba as u8 ptr, byval width as s32, byval height as s32, byval blocks as const any ptr, byval flags as s32)

	' maximum texture size is 2048 x 2048 texel.
	#define SQUISH_MAX_IMAGE_SIZE	2048 * 2048 * 4

	' squish.image_rgba() and squish.dxtblocks() are allocated as fixed size static array.
	dim shared metric_uniform(0 to 2) as float = {1.0, 1.0, 1.0}
	dim shared metric_perceptual(0 to 2) as float = {0.2126, 0.7152, 0.0722}
	dim shared image_rgba(0 to SQUISH_MAX_IMAGE_SIZE - 1) as u8
	dim shared effect_rgba(0 to SQUISH_MAX_IMAGE_SIZE - 1) as u8
	dim shared dxtblocks(0 to SQUISH_MAX_IMAGE_SIZE  - 1) as u8

	end namespace
	
end extern