'----------------------------------------------------------
'	Simple XML format support 
'	Copyright (c) 2015 Masamichi Kirikae
'----------------------------------------------------------

#include once "xml.bi"

namespace xml

sub push(byref st as string)
	xml_stack.depth += 1
	xml_stack.ename(xml_stack.depth) = st
end sub

function pop() as string
	dim st as string
	st = xml_stack.ename(xml_stack.depth)
	xml_stack.depth -= 1
	return(st)
end function

sub new_element(byref ename as string)
	dim i as s32
	push(ename)
	if (xml_stack.depth > 1) then
		for i = 1 to xml_stack.depth - 1
			print #xml_fh, chr(9);	' print tab
		next
	end if
	print #xml_fh, "<"; ename; ">"
end sub

sub close_element()
	dim i as s32, ename as string
	if (xml_stack.depth > 1) then
		for i = 1 to xml_stack.depth - 1
			print #xml_fh, chr(9);
		next
	end if
	ename = pop()
	print #xml_fh, "</"; ename; ">"
end sub

sub write_single_element(byref ename as string, byref value as string)
	dim i as s32
	for i = 1 to xml_stack.depth
		print #xml_fh, chr(9);
	next
	print #xml_fh, "<"; ename; ">"; value; "</"; ename; ">"	
end sub

function new_doc(byref filename as string, byref doctype as string) as s32
	xml_fh = freefile
	open filename for output as #xml_fh
	if (err > 1) then
		return(XML_ERROR_FILE_HANDLE)
	endif
	xml_stack.depth = 0
	print #xml_fh, xml_header
	new_element(doctype)
	return(XML_NO_ERROR)
end function

function close_doc() as s32
	if (xml_fh <> null) then
		close_element()
		close #xml_fh
		return(XML_NO_ERROR)
	else
		return(XML_ERROR_FILE_HANDLE)
	endif	
end function

function read_doc(byref filename as string) as s32
	dim as string st, doctype
	dim as s32 result
	xml_fh = freefile
	open filename for input as #xml_fh
	if (err > 1) then
		cprint using "ERROR:Can not open &."; filename
		return(XML_ERROR_FILE_HANDLE)
	endif
	xml_stack.depth = 0
	st = libfio.fgetstring(xml_fh, 39)
	if (len(st) = 39) then
		fpseek(xml_fh, 0)
		line input #xml_fh, st
		if (st <> xml_header) then
			cprint "ERROR:XML header not found."
			close #xml_fh
			return(XML_ERROR_SYNTAX)
		endif
	else
		cprint "ERROR:This file is not a XML document."
		close #xml_fh
		return(XML_ERROR_SYNTAX)
	endif
	return(XML_NO_ERROR)
end function

function get_element_status(byref st as string) as s32
	dim as s32 counter, i, lg, flag
	dim mst as string

	flag = false
	counter = 0
	lg = len(st)
	for i = 1 to lg
		mst = mid(st, i, 1)
		if (flag = false) then
			select case mst
				case "/", "-"
					return(XML_NO_ELEMENT)	'	this is a comment line
				case else
			end select
		endif
		select case mst
			case "<"
				if (flag = false) then
					flag = true
					counter+=1
				else
					counter+=1
				endif
			case ">"
				if (flag = true) then
					counter+=1
				end if
			case else
		end select
	next
	select case counter
		case 0
			return(XML_NO_ELEMENT)
		case 2
			return(XML_ELEMENT_ONLY)
		case 4
			return(XML_ELEMENT_WITH_VALUE)
		case else
			return(XML_ERROR_SYNTAX)
	end select
end function

sub get_element_name(byref st as string, byref ename as string)
	dim as s32 i, lg, flag
	dim as string*1 mst
	
	lg = len(st)
	ename = ""
	flag = false
	for i = 1 to lg
		mst = mid(st, i, 1)
		select case mst
			case "<"
				flag = true
			case ">"
				exit for
			case else
				if (flag = true) then
					ename = ename + mst
				end if
		end select
	next
end sub

sub get_element_and_value(byref st as string, byref ename as string, byref value as string)
	dim as s32 i, j, lg, counter
	dim as string*1 mst

	lg = len(st)
	ename = ""
	value = ""
	counter = 0
	for i = 1 to lg
		mst = mid(st, i, 1)
		select case mst
			case "<"
				counter += 1
			case ">"
				counter += 1
			case else
				select case counter
					case 0
						'	nop
					case 1
						ename = ename + mst
					case 2
						value = value + mst
					case else
						'	nop
				end select
		end select
	next
end sub

function read_element(byref ename as string, byref value as string) as s32
	dim st as string, flag as s32
	do until eof(xml_fh)
		line input #xml_fh, st
		flag = get_element_status(st)
		select case flag
			case XML_ELEMENT_ONLY
				get_element_name(st, ename)
				exit do
			case XML_ELEMENT_WITH_VALUE
				get_element_and_value(st, ename, value)
				exit do
			case XML_ERROR_SYNTAX
				close #xml_fh
				return(XML_ERROR_SYNTAX)
				exit do
			case XML_NO_ELEMENT	' comment line or blank line
				'	NOP
		end select
	loop
	return(flag)
end function

end namespace
